#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:softtabstop=4:smarttab:expandtab

from __future__ import unicode_literals, division, absolute_import, print_function

from .compatibility_utils import PY2, hexlify, bstr, bord, bchar
from loguru import logger

import datetime

if PY2:
    range = xrange

# note:  struct pack, unpack, unpack_from all require bytestring format
# data all the way up to at least python 2.7.5, python 3 okay with bytestring
import struct

from .unipath import pathof

DUMP = False
""" Set to True to dump all possible information. """


class unpackException(Exception):
    pass


def describe(data):
    txtans = ""
    hexans = hexlify(data)
    for i in data:
        if bord(i) < 32 or bord(i) > 127:
            txtans += "?"
        else:
            txtans += bchar(i).decode("latin-1")
    return '"' + txtans + '"' + " 0x" + hexans


def datetimefrompalmtime(palmtime):
    if palmtime > 0x7FFFFFFF:
        pythondatetime = datetime.datetime(
            year=1904, month=1, day=1
        ) + datetime.timedelta(seconds=palmtime)
    else:
        pythondatetime = datetime.datetime(
            year=1970, month=1, day=1
        ) + datetime.timedelta(seconds=palmtime)
    return pythondatetime


class Sectionizer:
    def __init__(self, filename):
        self.data = b""
        with open(pathof(filename), "rb") as f:
            self.data = f.read()
        self.palmheader = self.data[:78]
        self.palmname = self.data[:32]
        self.ident = self.palmheader[0x3C : 0x3C + 8]
        (self.num_sections,) = struct.unpack_from(b">H", self.palmheader, 76)
        self.filelength = len(self.data)
        sectionsdata = struct.unpack_from(
            bstr(">%dL" % (self.num_sections * 2)), self.data, 78
        ) + (self.filelength, 0)
        self.sectionoffsets = sectionsdata[::2]
        self.sectionattributes = sectionsdata[1::2]
        self.sectiondescriptions = ["" for x in range(self.num_sections + 1)]
        self.sectiondescriptions[-1] = "File Length Only"
        return

    def dumpsectionsinfo(self):
        logger.debug("Section     Offset  Length      UID Attribs Description")
        for i in range(self.num_sections):
            logger.debug(
                "%3d %3X  0x%07X 0x%05X % 8d % 7d %s"
                % (
                    i,
                    i,
                    self.sectionoffsets[i],
                    self.sectionoffsets[i + 1] - self.sectionoffsets[i],
                    self.sectionattributes[i] & 0xFFFFFF,
                    (self.sectionattributes[i] >> 24) & 0xFF,
                    self.sectiondescriptions[i],
                )
            )
        logger.debug(
            "%3d %3X  0x%07X                          %s"
            % (
                self.num_sections,
                self.num_sections,
                self.sectionoffsets[self.num_sections],
                self.sectiondescriptions[self.num_sections],
            )
        )

    def setsectiondescription(self, section, description):
        if section < len(self.sectiondescriptions):
            self.sectiondescriptions[section] = description
        else:
            logger.debug(
                "Section out of range: %d, description %s" % (section, description)
            )

    def dumppalmheader(self):
        logger.debug("Palm Database Header")
        logger.debug("Database name: " + repr(self.palmheader[:32]))
        (dbattributes,) = struct.unpack_from(b">H", self.palmheader, 32)
        logger.debug("Bitfield attributes: 0x%0X" % dbattributes,)
        if dbattributes != 0:
            print(" (",)
            if dbattributes & 2:
                print("Read-only; ",)
            if dbattributes & 4:
                print("Dirty AppInfoArea; ",)
            if dbattributes & 8:
                print("Needs to be backed up; ",)
            if dbattributes & 16:
                print("OK to install over newer; ",)
            if dbattributes & 32:
                print("Reset after installation; ",)
            if dbattributes & 64:
                print("No copying by PalmPilot beaming; ",)
            print(")")
        else:
            print("")
        logger.debug(
            "File version: %d" % struct.unpack_from(b">H", self.palmheader, 34)[0]
        )
        (dbcreation,) = struct.unpack_from(b">L", self.palmheader, 36)
        logger.debug(
            "Creation Date: "
            + str(datetimefrompalmtime(dbcreation))
            + (" (0x%0X)" % dbcreation)
        )
        (dbmodification,) = struct.unpack_from(b">L", self.palmheader, 40)
        logger.debug(
            "Modification Date: "
            + str(datetimefrompalmtime(dbmodification))
            + (" (0x%0X)" % dbmodification)
        )
        (dbbackup,) = struct.unpack_from(b">L", self.palmheader, 44)
        if dbbackup != 0:
            logger.debug(
                "Backup Date: "
                + str(datetimefrompalmtime(dbbackup))
                + (" (0x%0X)" % dbbackup)
            )
        logger.debug(
            "Modification No.: %d" % struct.unpack_from(b">L", self.palmheader, 48)[0]
        )
        logger.debug(
            "App Info offset: 0x%0X" % struct.unpack_from(b">L", self.palmheader, 52)[0]
        )
        logger.debug(
            "Sort Info offset: 0x%0X"
            % struct.unpack_from(b">L", self.palmheader, 56)[0]
        )
        logger.debug(
            "Type/Creator: %s/%s"
            % (repr(self.palmheader[60:64]), repr(self.palmheader[64:68]))
        )
        logger.debug(
            "Unique seed: 0x%0X" % struct.unpack_from(b">L", self.palmheader, 68)[0]
        )
        (expectedzero,) = struct.unpack_from(b">L", self.palmheader, 72)
        if expectedzero != 0:
            logger.debug(
                "Should be zero but isn't: %d"
                % struct.unpack_from(b">L", self.palmheader, 72)[0]
            )
        logger.debug(
            "Number of sections: %d" % struct.unpack_from(b">H", self.palmheader, 76)[0]
        )
        return

    def loadSection(self, section):
        before, after = self.sectionoffsets[section : section + 2]
        return self.data[before:after]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:softtabstop=4:smarttab:expandtab

from __future__ import unicode_literals, division, absolute_import, print_function

from .compatibility_utils import unicode_str
import os
from .unipath import pathof
from loguru import logger

import re

# note: re requites the pattern to be the exact same type as the data to be searched in python3
# but u"" is not allowed for the pattern itself only b""

DEBUG_NAV = False

FORCE_DEFAULT_TITLE = False
""" Set to True to force to use the default title. """

NAVIGATION_FINENAME = "nav.xhtml"
""" The name for the navigation document. """

DEFAULT_TITLE = "Navigation"
""" The default title for the navigation document. """


class NAVProcessor(object):
    def __init__(self, files):
        self.files = files
        self.navname = NAVIGATION_FINENAME

    def buildLandmarks(self, guidetext):
        header = ""
        header += '  <nav epub:type="landmarks" id="landmarks" hidden="">\n'
        header += "    <h2>Guide</h2>\n"
        header += "    <ol>\n"
        element = '      <li><a epub:type="{:s}" href="{:s}">{:s}</a></li>\n'
        footer = ""
        footer += "    </ol>\n"
        footer += "  </nav>\n"

        type_map = {
            "cover": "cover",
            "title-page": "title-page",
            # ?: 'frontmatter',
            "text": "bodymatter",
            # ?: 'backmatter',
            "toc": "toc",
            "loi": "loi",
            "lot": "lot",
            "preface": "preface",
            "bibliography": "bibliography",
            "index": "index",
            "glossary": "glossary",
            "acknowledgements": "acknowledgements",
            "colophon": None,
            "copyright-page": None,
            "dedication": None,
            "epigraph": None,
            "foreword": None,
            "notes": None,
        }

        re_type = re.compile(r'\s+type\s*=\s*"(.*?)"', re.I)
        re_title = re.compile(r'\s+title\s*=\s*"(.*?)"', re.I)
        re_link = re.compile(r'\s+href\s*=\s*"(.*?)"', re.I)
        dir_ = os.path.relpath(self.files.k8text, self.files.k8oebps).replace("\\", "/")

        data = ""
        references = re.findall(r"<reference\s+.*?>", unicode_str(guidetext), re.I)
        for reference in references:
            mo_type = re_type.search(reference)
            mo_title = re_title.search(reference)
            mo_link = re_link.search(reference)
            if mo_type is not None:
                type_ = type_map.get(mo_type.group(1), None)
            else:
                type_ = None
            if mo_title is not None:
                title = mo_title.group(1)
            else:
                title = None
            if mo_link is not None:
                link = mo_link.group(1)
            else:
                link = None

            if type_ is not None and title is not None and link is not None:
                link = os.path.relpath(link, dir_).replace("\\", "/")
                data += element.format(type_, link, title)
        if len(data) > 0:
            return header + data + footer
        else:
            return ""

    def buildTOC(self, indx_data):
        header = ""
        header += '  <nav epub:type="toc" id="toc">\n'
        header += "    <h1>Table of contents</h1>\n"
        footer = "  </nav>\n"

        # recursive part
        def recursINDX(max_lvl=0, num=0, lvl=0, start=-1, end=-1):
            if start > len(indx_data) or end > len(indx_data):
                logger.debug(
                    "Warning (in buildTOC): missing INDX child entries",
                    start,
                    end,
                    len(indx_data),
                )
                return ""
            if DEBUG_NAV:
                logger.debug(
                    "recursINDX (in buildTOC) lvl %d from %d to %d" % (lvl, start, end)
                )
            xhtml = ""
            if start <= 0:
                start = 0
            if end <= 0:
                end = len(indx_data)
            if lvl > max_lvl:
                max_lvl = lvl

            indent1 = "  " * (2 + lvl * 2)
            indent2 = "  " * (3 + lvl * 2)
            xhtml += indent1 + "<ol>\n"
            for i in range(start, end):
                e = indx_data[i]
                htmlfile = e["filename"]
                desttag = e["idtag"]
                text = e["text"]
                if not e["hlvl"] == lvl:
                    continue
                num += 1
                if desttag == "":
                    link = htmlfile
                else:
                    link = "{:s}#{:s}".format(htmlfile, desttag)
                xhtml += indent2 + "<li>"
                entry = '<a href="{:}">{:s}</a>'.format(link, text)
                xhtml += entry
                # recurs
                if e["child1"] >= 0:
                    xhtml += "\n"
                    xhtmlrec, max_lvl, num = recursINDX(
                        max_lvl, num, lvl + 1, e["child1"], e["childn"] + 1
                    )
                    xhtml += xhtmlrec
                    xhtml += indent2
                # close entry
                xhtml += "</li>\n"
            xhtml += indent1 + "</ol>\n"
            return xhtml, max_lvl, num

        data, max_lvl, num = recursINDX()
        if not len(indx_data) == num:
            logger.debug(
                "Warning (in buildTOC): different number of entries in NCX",
                len(indx_data),
                num,
            )
        return header + data + footer

    def buildNAV(self, ncx_data, guidetext, title, lang):
        logger.debug("Building Navigation Document.")
        if FORCE_DEFAULT_TITLE:
            title = DEFAULT_TITLE
        nav_header = ""
        nav_header += '<?xml version="1.0" encoding="utf-8"?>\n<!DOCTYPE html>'
        nav_header += '<html xmlns="http://www.w3.org/1999/xhtml"'
        nav_header += ' xmlns:epub="http://www.idpf.org/2007/ops"'
        nav_header += ' lang="{0:s}" xml:lang="{0:s}">\n'.format(lang)
        nav_header += "<head>\n<title>{:s}</title>\n".format(title)
        nav_header += '<meta charset="UTF-8" />\n'
        nav_header += '<style type="text/css">\n'
        nav_header += "nav#landmarks { display:none; }\n"
        nav_header += "</style>\n</head>\n<body>\n"
        nav_footer = "</body>\n</html>\n"

        landmarks = self.buildLandmarks(guidetext)
        toc = self.buildTOC(ncx_data)

        data = nav_header
        data += landmarks
        data += toc
        data += nav_footer
        return data

    def getNAVName(self):
        return self.navname

    def writeNAV(self, ncx_data, guidetext, metadata):
        # build the xhtml
        # logger.debug("Write Navigation Document.")
        xhtml = self.buildNAV(
            ncx_data, guidetext, metadata.get("Title")[0], metadata.get("Language")[0]
        )
        fname = os.path.join(self.files.k8text, self.navname)
        with open(pathof(fname), "wb") as f:
            f.write(xhtml.encode("utf-8"))



#! /usr/bin/python
# vim:ts=4:sw=4:softtabstop=4:smarttab:expandtab


# this program works in concert with the output from KindleUnpack

"""
Convert from Mobi ML to XHTML
"""

import os
import sys
import re

SPECIAL_HANDLING_TAGS = {
    "?xml": ("xmlheader", -1),
    "!--": ("comment", -3),
    "!DOCTYPE": ("doctype", -1),
}

SPECIAL_HANDLING_TYPES = ["xmlheader", "doctype", "comment"]

SELF_CLOSING_TAGS = [
    "br",
    "hr",
    "input",
    "img",
    "image",
    "meta",
    "spacer",
    "link",
    "frame",
    "base",
    "col",
    "reference",
]


class MobiMLConverter(object):

    PAGE_BREAK_PAT = re.compile(r"(<[/]{0,1}mbp:pagebreak\s*[/]{0,1}>)+", re.IGNORECASE)
    IMAGE_ATTRS = ("lowrecindex", "recindex", "hirecindex")

    def __init__(self, filename):
        self.base_css_rules = "blockquote { margin: 0em 0em 0em 1.25em }\n"
        self.base_css_rules += "p { margin: 0em }\n"
        self.base_css_rules += ".bold { font-weight: bold }\n"
        self.base_css_rules += ".italic { font-style: italic }\n"
        self.base_css_rules += (
            ".mbp_pagebreak { page-break-after: always; margin: 0; display: block }\n"
        )
        self.tag_css_rules = {}
        self.tag_css_rule_cnt = 0
        self.path = []
        self.filename = filename
        self.wipml = open(self.filename, "rb").read()
        self.pos = 0
        self.opfname = self.filename.rsplit(".", 1)[0] + ".opf"
        self.opos = 0
        self.meta = ""
        self.cssname = os.path.join(os.path.dirname(self.filename), "styles.css")
        self.current_font_size = 3
        self.font_history = []

    def cleanup_html(self):
        self.wipml = re.sub(
            r'<div height="0(pt|px|ex|em|%){0,1}"></div>', "", self.wipml
        )
        self.wipml = self.wipml.replace("\r\n", "\n")
        self.wipml = self.wipml.replace("> <", ">\n<")
        self.wipml = self.wipml.replace("<mbp: ", "<mbp:")
        # self.wipml = re.sub(r'<?xml[^>]*>', '', self.wipml)
        self.wipml = self.wipml.replace("<br></br>", "<br/>")

    def replace_page_breaks(self):
        self.wipml = self.PAGE_BREAK_PAT.sub(
            '<div class="mbp_pagebreak" />', self.wipml
        )

    # parse leading text of ml and tag
    def parseml(self):
        p = self.pos
        if p >= len(self.wipml):
            return None
        if self.wipml[p] != "<":
            res = self.wipml.find("<", p)
            if res == -1:
                res = len(self.wipml)
            self.pos = res
            return self.wipml[p:res], None
        # handle comment as a special case to deal with multi-line comments
        if self.wipml[p : p + 4] == "<!--":
            te = self.wipml.find("-->", p + 1)
            if te != -1:
                te = te + 2
        else:
            te = self.wipml.find(">", p + 1)
            ntb = self.wipml.find("<", p + 1)
            if ntb != -1 and ntb < te:
                self.pos = ntb
                return self.wipml[p:ntb], None
        self.pos = te + 1
        return None, self.wipml[p : te + 1]

    # parses string version of tag to identify its name,
    # its type 'begin', 'end' or 'single',
    # plus build a hashtable of its attributes
    # code is written to handle the possiblity of very poor formating
    def parsetag(self, s):
        p = 1
        # get the tag name
        tname = None
        ttype = None
        tattr = {}
        while s[p : p + 1] == " ":
            p += 1
        if s[p : p + 1] == "/":
            ttype = "end"
            p += 1
            while s[p : p + 1] == " ":
                p += 1
        b = p
        while s[p : p + 1] not in (">", "/", " ", '"', "'", "\r", "\n"):
            p += 1
        tname = s[b:p].lower()
        if tname == "!doctype":
            tname = "!DOCTYPE"
        # special cases
        if tname in SPECIAL_HANDLING_TAGS.keys():
            ttype, backstep = SPECIAL_HANDLING_TAGS[tname]
            tattr["special"] = s[p:backstep]
        if ttype is None:
            # parse any attributes
            while s.find("=", p) != -1:
                while s[p : p + 1] == " ":
                    p += 1
                b = p
                while s[p : p + 1] != "=":
                    p += 1
                aname = s[b:p].lower()
                aname = aname.rstrip(" ")
                p += 1
                while s[p : p + 1] == " ":
                    p += 1
                if s[p : p + 1] in ('"', "'"):
                    p = p + 1
                    b = p
                    while s[p : p + 1] not in ('"', "'"):
                        p += 1
                    val = s[b:p]
                    p += 1
                else:
                    b = p
                    while s[p : p + 1] not in (">", "/", " "):
                        p += 1
                    val = s[b:p]
                tattr[aname] = val
        # label beginning and single tags
        if ttype is None:
            ttype = "begin"
            if s.find(" /", p) >= 0:
                ttype = "single_ext"
            elif s.find("/", p) >= 0:
                ttype = "single"
        return ttype, tname, tattr

    # main routine to convert from mobi markup language to html
    def processml(self):

        # are these really needed
        html_done = False
        head_done = False
        body_done = False

        skip = False

        htmlstr = ""
        self.replace_page_breaks()
        self.cleanup_html()

        # now parse the cleaned up ml into standard xhtml
        while True:

            r = self.parseml()
            if not r:
                break

            text, tag = r

            if text:
                if not skip:
                    htmlstr += text

            if tag:
                ttype, tname, tattr = self.parsetag(tag)

                # If we run into a DTD or xml declarations inside the body ... bail.
                if (
                    tname in SPECIAL_HANDLING_TAGS.keys()
                    and tname != "comment"
                    and body_done
                ):
                    htmlstr += "\n</body></html>"
                    break

                # make sure self-closing tags actually self-close
                if ttype == "begin" and tname in SELF_CLOSING_TAGS:
                    ttype = "single"

                # make sure any end tags of self-closing tags are discarded
                if ttype == "end" and tname in SELF_CLOSING_TAGS:
                    continue

                # remove embedded guide and refernces from old mobis
                if tname in ("guide", "ncx", "reference") and ttype in (
                    "begin",
                    "single",
                    "single_ext",
                ):
                    tname = "removeme:{0}".format(tname)
                    tattr = None
                if (
                    tname in ("guide", "ncx", "reference", "font", "span")
                    and ttype == "end"
                ):
                    if self.path[-1] == "removeme:{0}".format(tname):
                        tname = "removeme:{0}".format(tname)
                        tattr = None

                # Get rid of font tags that only have a color attribute.
                if tname == "font" and ttype in ("begin", "single", "single_ext"):
                    if "color" in tattr.keys() and len(tattr.keys()) == 1:
                        tname = "removeme:{0}".format(tname)
                        tattr = None

                # Get rid of empty spans in the markup.
                if (
                    tname == "span"
                    and ttype in ("begin", "single", "single_ext")
                    and not len(tattr)
                ):
                    tname = "removeme:{0}".format(tname)

                # need to handle fonts outside of the normal methods
                # so fonts tags won't be added to the self.path since we keep track
                # of font tags separately with self.font_history
                if tname == "font" and ttype == "begin":
                    # check for nested font start tags
                    if len(self.font_history) > 0:
                        # inject a font end tag
                        taginfo = ("end", "font", None)
                        htmlstr += self.processtag(taginfo)
                    self.font_history.append((ttype, tname, tattr))
                    # handle the current font start tag
                    taginfo = (ttype, tname, tattr)
                    htmlstr += self.processtag(taginfo)
                    continue

                # check for nested font tags and unnest them
                if tname == "font" and ttype == "end":
                    self.font_history.pop()
                    # handle this font end tag
                    taginfo = ("end", "font", None)
                    htmlstr += self.processtag(taginfo)
                    # check if we were nested
                    if len(self.font_history) > 0:
                        # inject a copy of the most recent font start tag from history
                        taginfo = self.font_history[-1]
                        htmlstr += self.processtag(taginfo)
                    continue

                # keep track of nesting path
                if ttype == "begin":
                    self.path.append(tname)
                elif ttype == "end":
                    if tname != self.path[-1]:
                        print ("improper nesting: ", self.path, tname, ttype)
                        if tname not in self.path:
                            # handle case of end tag with no beginning by injecting empty begin tag
                            taginfo = ("begin", tname, None)
                            htmlstr += self.processtag(taginfo)
                            print "     - fixed by injecting empty start tag ", tname
                            self.path.append(tname)
                        elif len(self.path) > 1 and tname == self.path[-2]:
                            # handle case of dangling missing end
                            taginfo = ("end", self.path[-1], None)
                            htmlstr += self.processtag(taginfo)
                            print "     - fixed by injecting end tag ", self.path[-1]
                            self.path.pop()
                    self.path.pop()

                if tname == "removeme:{0}".format(tname):
                    if ttype in ("begin", "single", "single_ext"):
                        skip = True
                    else:
                        skip = False
                else:
                    taginfo = (ttype, tname, tattr)
                    htmlstr += self.processtag(taginfo)

                # handle potential issue of multiple html, head, and body sections
                if tname == "html" and ttype == "begin" and not html_done:
                    htmlstr += "\n"
                    html_done = True

                if tname == "head" and ttype == "begin" and not head_done:
                    htmlstr += "\n"
                    # also add in metadata and style link tags
                    htmlstr += self.meta
                    htmlstr += (
                        '<link href="styles.css" rel="stylesheet" type="text/css" />\n'
                    )
                    head_done = True

                if tname == "body" and ttype == "begin" and not body_done:
                    htmlstr += "\n"
                    body_done = True

        # handle issue of possibly missing html, head, and body tags
        # I have not seen this but the original did something like this so ...
        if not body_done:
            htmlstr = "<body>\n" + htmlstr + "</body>\n"
        if not head_done:
            headstr = "<head>\n"
            headstr += self.meta
            headstr += '<link href="styles.css" rel="stylesheet" type="text/css" />\n'
            headstr += "</head>\n"
            htmlstr = headstr + htmlstr
        if not html_done:
            htmlstr = "<html>\n" + htmlstr + "</html>\n"

        # finally add DOCTYPE info
        htmlstr = (
            '<?xml version="1.0"?>\n<!DOCTYPE HTML PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">\n'
            + htmlstr
        )

        css = self.base_css_rules
        for cls, rule in self.tag_css_rules.items():
            css += ".%s { %s }\n" % (cls, rule)

        return (htmlstr, css, self.cssname)

    def ensure_unit(self, raw, unit="px"):
        if re.search(r"\d+$", raw) is not None:
            raw += unit
        return raw

    # flatten possibly modified tag back to string
    def taginfo_tostring(self, taginfo):
        (ttype, tname, tattr) = taginfo
        if ttype is None or tname is None:
            return ""
        if ttype == "end":
            return "</%s>" % tname
        if (
            ttype in SPECIAL_HANDLING_TYPES
            and tattr is not None
            and "special" in tattr.keys()
        ):
            info = tattr["special"]
            if ttype == "comment":
                return "<%s %s-->" % tname, info
            else:
                return "<%s %s>" % tname, info
        res = []
        res.append("<%s" % tname)
        if tattr is not None:
            for key in tattr.keys():
                res.append(' %s="%s"' % (key, tattr[key]))
        if ttype == "single":
            res.append("/>")
        elif ttype == "single_ext":
            res.append(" />")
        else:
            res.append(">")
        return "".join(res)

    # routines to convert from mobi ml tags atributes to xhtml attributes and styles
    def processtag(self, taginfo):
        # Converting mobi font sizes to numerics
        size_map = {
            "xx-small": "1",
            "x-small": "2",
            "small": "3",
            "medium": "4",
            "large": "5",
            "x-large": "6",
            "xx-large": "7",
        }

        size_to_em_map = {
            "1": ".65em",
            "2": ".75em",
            "3": "1em",
            "4": "1.125em",
            "5": "1.25em",
            "6": "1.5em",
            "7": "2em",
        }

        # current tag to work on
        (ttype, tname, tattr) = taginfo
        if not tattr:
            tattr = {}

        styles = []

        if tname is None or tname.startswith("removeme"):
            return ""

        # have not seen an example of this yet so keep it here to be safe
        # until this is better understood
        if tname in (
            "country-region",
            "place",
            "placetype",
            "placename",
            "state",
            "city",
            "street",
            "address",
            "content",
        ):
            tname = "div" if tname == "content" else "span"
            for key in tattr.keys():
                tattr.pop(key)

        # handle general case of style, height, width, bgcolor in any tag
        if "style" in tattr.keys():
            style = tattr.pop("style").strip()
            if style:
                styles.append(style)

        if "align" in tattr.keys():
            align = tattr.pop("align").strip()
            if align:
                if tname in ("table", "td", "tr"):
                    pass
                else:
                    styles.append("text-align: %s" % align)

        if "height" in tattr.keys():
            height = tattr.pop("height").strip()
            if (
                height
                and "<" not in height
                and ">" not in height
                and re.search(r"\d+", height)
            ):
                if tname in ("table", "td", "tr"):
                    pass
                elif tname == "img":
                    tattr["height"] = height
                else:
                    styles.append("margin-top: %s" % self.ensure_unit(height))

        if "width" in tattr.keys():
            width = tattr.pop("width").strip()
            if width and re.search(r"\d+", width):
                if tname in ("table", "td", "tr"):
                    pass
                elif tname == "img":
                    tattr["width"] = width
                else:
                    styles.append("text-indent: %s" % self.ensure_unit(width))
                    if width.startswith("-"):
                        styles.append("margin-left: %s" % self.ensure_unit(width[1:]))

        if "bgcolor" in tattr.keys():
            # no proprietary html allowed
            if tname == "div":
                del tattr["bgcolor"]

        elif tname == "font":
            # Change font tags to span tags
            tname = "span"
            if ttype in ("begin", "single", "single_ext"):
                # move the face attribute to css font-family
                if "face" in tattr.keys():
                    face = tattr.pop("face").strip()
                    styles.append('font-family: "%s"' % face)

                    # Monitor the constantly changing font sizes, change them to ems and move
                    # them to css. The following will work for 'flat' font tags, but nested font tags
                    # will cause things to go wonky. Need to revert to the parent font tag's size
                    # when a closing tag is encountered.
                if "size" in tattr.keys():
                    sz = tattr.pop("size").strip().lower()
                    try:
                        float(sz)
                    except ValueError:
                        if sz in size_map.keys():
                            sz = size_map[sz]
                    else:
                        if sz.startswith("-") or sz.startswith("+"):
                            sz = self.current_font_size + float(sz)
                            if sz > 7:
                                sz = 7
                            elif sz < 1:
                                sz = 1
                            sz = str(int(sz))
                    styles.append("font-size: %s" % size_to_em_map[sz])
                    self.current_font_size = int(sz)

        elif tname == "img":
            for attr in ("width", "height"):
                if attr in tattr:
                    val = tattr[attr]
                    if val.lower().endswith("em"):
                        try:
                            nval = float(val[:-2])
                            nval *= 16 * (
                                168.451 / 72
                            )  # Assume this was set using the Kindle profile
                            tattr[attr] = "%dpx" % int(nval)
                        except:
                            del tattr[attr]
                    elif val.lower().endswith("%"):
                        del tattr[attr]

        # convert the anchor tags
        if "filepos-id" in tattr:
            tattr["id"] = tattr.pop("filepos-id")
            if "name" in tattr and tattr["name"] != tattr["id"]:
                tattr["name"] = tattr["id"]

        if "filepos" in tattr:
            filepos = tattr.pop("filepos")
            try:
                tattr["href"] = "#filepos%d" % int(filepos)
            except ValueError:
                pass

        if styles:
            ncls = None
            rule = "; ".join(styles)
            for sel, srule in self.tag_css_rules.items():
                if srule == rule:
                    ncls = sel
                    break
            if ncls is None:
                self.tag_css_rule_cnt += 1
                ncls = "rule_%d" % self.tag_css_rule_cnt
                self.tag_css_rules[ncls] = rule
            cls = tattr.get("class", "")
            cls = cls + (" " if cls else "") + ncls
            tattr["class"] = cls

        # convert updated tag back to string representation
        if len(tattr) == 0:
            tattr = None
        taginfo = (ttype, tname, tattr)
        return self.taginfo_tostring(taginfo)


""" main only left in for testing outside of plugin """


def main(argv=sys.argv):
    if len(argv) != 2:
        return 1
    else:
        infile = argv[1]

    try:
        print "Converting Mobi Markup Language to XHTML"
        mlc = MobiMLConverter(infile)
        print "Processing ..."
        htmlstr, css, cssname = mlc.processml()
        outname = infile.rsplit(".", 1)[0] + "_converted.html"
        file(outname, "wb").write(htmlstr)
        file(cssname, "wb").write(css)
        print "Completed"
        print "XHTML version of book can be found at: " + outname

    except ValueError, e:
        print "Error: %s" % e
        return 1

    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:softtabstop=4:smarttab:expandtab
# flake8: noqa

from __future__ import unicode_literals, division, absolute_import, print_function

from .compatibility_utils import PY2, text_type, bchr, bord

import binascii

if PY2:
    range = xrange

from itertools import cycle


def getLanguage(langID, sublangID):
    mobilangdict = {
        54: {0: "af"},  # Afrikaans
        28: {0: "sq"},  # Albanian
        1: {
            0: "ar",
            5: "ar-dz",
            15: "ar-bh",
            3: "ar-eg",
            2: "ar-iq",
            11: "ar-jo",
            13: "ar-kw",
            12: "ar-lb",
            4: "ar-ly",
            6: "ar-ma",
            8: "ar-om",
            16: "ar-qa",
            1: "ar-sa",
            10: "ar-sy",
            7: "ar-tn",
            14: "ar-ae",
            9: "ar-ye",
        },
        # Arabic,  Arabic (Algeria),  Arabic (Bahrain),  Arabic (Egypt),  Arabic
        # (Iraq), Arabic (Jordan),  Arabic (Kuwait),  Arabic (Lebanon),  Arabic
        # (Libya), Arabic (Morocco),  Arabic (Oman),  Arabic (Qatar),  Arabic
        # (Saudi Arabia),  Arabic (Syria),  Arabic (Tunisia),  Arabic (United Arab
        # Emirates),  Arabic (Yemen)
        43: {0: "hy"},  # Armenian
        77: {0: "as"},  # Assamese
        44: {0: "az"},  # "Azeri (IANA: Azerbaijani)
        45: {0: "eu"},  # Basque
        35: {0: "be"},  # Belarusian
        69: {0: "bn"},  # Bengali
        2: {0: "bg"},  # Bulgarian
        3: {0: "ca"},  # Catalan
        4: {0: "zh", 3: "zh-hk", 2: "zh-cn", 4: "zh-sg", 1: "zh-tw"},
        # Chinese,  Chinese (Hong Kong),  Chinese (PRC),  Chinese (Singapore),  Chinese (Taiwan)
        26: {0: "hr", 3: "sr"},  # Croatian, Serbian
        5: {0: "cs"},  # Czech
        6: {0: "da"},  # Danish
        19: {0: "nl", 1: "nl", 2: "nl-be"},  # Dutch / Flemish,  Dutch (Belgium)
        9: {
            0: "en",
            1: "en",
            3: "en-au",
            40: "en-bz",
            4: "en-ca",
            6: "en-ie",
            8: "en-jm",
            5: "en-nz",
            13: "en-ph",
            7: "en-za",
            11: "en-tt",
            2: "en-gb",
            1: "en-us",
            12: "en-zw",
        },
        # English,  English (Australia),  English (Belize),  English (Canada),
        # English (Ireland),  English (Jamaica),  English (New Zealand),  English
        # (Philippines),  English (South Africa),  English (Trinidad),  English
        # (United Kingdom),  English (United States),  English (Zimbabwe)
        37: {0: "et"},  # Estonian
        56: {0: "fo"},  # Faroese
        41: {0: "fa"},  # Farsi / Persian
        11: {0: "fi"},  # Finnish
        12: {
            0: "fr",
            1: "fr",
            2: "fr-be",
            3: "fr-ca",
            5: "fr-lu",
            6: "fr-mc",
            4: "fr-ch",
        },
        # French,  French (Belgium),  French (Canada),  French (Luxembourg),  French (Monaco),  French (Switzerland)
        55: {0: "ka"},  # Georgian
        7: {0: "de", 1: "de", 3: "de-at", 5: "de-li", 4: "de-lu", 2: "de-ch"},
        # German,  German (Austria),  German (Liechtenstein),  German (Luxembourg),  German (Switzerland)
        8: {0: "el"},  # Greek, Modern (1453-)
        71: {0: "gu"},  # Gujarati
        13: {0: "he"},  # Hebrew (also code 'iw'?)
        57: {0: "hi"},  # Hindi
        14: {0: "hu"},  # Hungarian
        15: {0: "is"},  # Icelandic
        33: {0: "id"},  # Indonesian
        16: {0: "it", 1: "it", 2: "it-ch"},  # Italian,  Italian (Switzerland)
        17: {0: "ja"},  # Japanese
        75: {0: "kn"},  # Kannada
        63: {0: "kk"},  # Kazakh
        87: {0: "x-kok"},  # Konkani (real language code is 'kok'?)
        18: {0: "ko"},  # Korean
        38: {0: "lv"},  # Latvian
        39: {0: "lt"},  # Lithuanian
        47: {0: "mk"},  # Macedonian
        62: {0: "ms"},  # Malay
        76: {0: "ml"},  # Malayalam
        58: {0: "mt"},  # Maltese
        78: {0: "mr"},  # Marathi
        97: {0: "ne"},  # Nepali
        20: {0: "no"},  # Norwegian
        72: {0: "or"},  # Oriya
        21: {0: "pl"},  # Polish
        22: {0: "pt", 2: "pt", 1: "pt-br"},  # Portuguese,  Portuguese (Brazil)
        70: {0: "pa"},  # Punjabi
        23: {0: "rm"},  # "Rhaeto-Romanic" (IANA: Romansh)
        24: {0: "ro"},  # Romanian
        25: {0: "ru"},  # Russian
        59: {0: "sz"},  # "Sami (Lappish)" (not an IANA language code)
        # IANA code for "Northern Sami" is 'se'
        # 'SZ' is the IANA region code for Swaziland
        79: {0: "sa"},  # Sanskrit
        27: {0: "sk"},  # Slovak
        36: {0: "sl"},  # Slovenian
        46: {0: "sb"},  # "Sorbian" (not an IANA language code)
        # 'SB' is IANA region code for 'Solomon Islands'
        # Lower Sorbian = 'dsb'
        # Upper Sorbian = 'hsb'
        # Sorbian Languages = 'wen'
        10: {
            0: "es",
            4: "es",
            44: "es-ar",
            64: "es-bo",
            52: "es-cl",
            36: "es-co",
            20: "es-cr",
            28: "es-do",
            48: "es-ec",
            68: "es-sv",
            16: "es-gt",
            72: "es-hn",
            8: "es-mx",
            76: "es-ni",
            24: "es-pa",
            60: "es-py",
            40: "es-pe",
            80: "es-pr",
            56: "es-uy",
            32: "es-ve",
        },
        # Spanish,  Spanish (Mobipocket bug?),  Spanish (Argentina),  Spanish
        # (Bolivia),  Spanish (Chile),  Spanish (Colombia),  Spanish (Costa Rica),
        # Spanish (Dominican Republic),  Spanish (Ecuador),  Spanish (El
        # Salvador),  Spanish (Guatemala),  Spanish (Honduras),  Spanish (Mexico),
        # Spanish (Nicaragua),  Spanish (Panama),  Spanish (Paraguay),  Spanish
        # (Peru),  Spanish (Puerto Rico),  Spanish (Uruguay),  Spanish (Venezuela)
        48: {0: "sx"},  # "Sutu" (not an IANA language code)
        # "Sutu" is another name for "Southern Sotho"?
        # IANA code for "Southern Sotho" is 'st'
        65: {0: "sw"},  # Swahili
        29: {0: "sv", 1: "sv", 8: "sv-fi"},  # Swedish,  Swedish (Finland)
        73: {0: "ta"},  # Tamil
        68: {0: "tt"},  # Tatar
        74: {0: "te"},  # Telugu
        30: {0: "th"},  # Thai
        49: {0: "ts"},  # Tsonga
        50: {0: "tn"},  # Tswana
        31: {0: "tr"},  # Turkish
        34: {0: "uk"},  # Ukrainian
        32: {0: "ur"},  # Urdu
        67: {0: "uz", 2: "uz"},  # Uzbek
        42: {0: "vi"},  # Vietnamese
        52: {0: "xh"},  # Xhosa
        53: {0: "zu"},  # Zulu
    }
    lang = "en"
    if langID in mobilangdict:
        subdict = mobilangdict[langID]
        lang = subdict[0]
        if sublangID in subdict:
            lang = subdict[sublangID]
    return lang


def toHex(byteList):
    return binascii.hexlify(byteList)


# returns base32 bytestring
def toBase32(value, npad=4):
    digits = b"0123456789ABCDEFGHIJKLMNOPQRSTUV"
    num_string = b""
    current = value
    while current != 0:
        next, remainder = divmod(current, 32)
        rem_string = digits[remainder : remainder + 1]
        num_string = rem_string + num_string
        current = next
    if num_string == b"":
        num_string = b"0"
    pad = npad - len(num_string)
    if pad > 0:
        num_string = b"0" * pad + num_string
    return num_string


# converts base32 string to value
def fromBase32(str_num):
    if isinstance(str_num, text_type):
        str_num = str_num.encode("latin-1")
    scalelst = [1, 32, 1024, 32768, 1048576, 33554432, 1073741824, 34359738368]
    value = 0
    j = 0
    n = len(str_num)
    scale = 0
    for i in range(n):
        c = str_num[n - i - 1 : n - i]
        if c in b"0123456789":
            v = ord(c) - ord(b"0")
        else:
            v = ord(c) - ord(b"A") + 10
        if j < len(scalelst):
            scale = scalelst[j]
        else:
            scale = scale * 32
        j += 1
        if v != 0:
            value = value + (v * scale)
    return value


# note: if decode a bytestring using 'latin-1' (or any other 0-255 encoding)
# in place of ascii you will get a byte to half-word or integer
# one to one mapping of values from 0 - 255


def mangle_fonts(encryption_key, data):
    if isinstance(encryption_key, text_type):
        encryption_key = encryption_key.encode("latin-1")
    crypt = data[:1024]
    key = cycle(iter(map(bord, encryption_key)))
    # encrypt = ''.join([chr(ord(x)^key.next()) for x in crypt])
    encrypt = b"".join([bchr(bord(x) ^ next(key)) for x in crypt])
    return encrypt + data[1024:]



# -*- coding: utf-8 -*-
import shutil

from loguru import logger
import tempfile
from os.path import basename, splitext, exists, join
from mobi.kindleunpack import unpackBook


def extract(infile):
    """Extract mobi file and return path to epub file"""

    tempdir = tempfile.mkdtemp(prefix="mobiex")
    if hasattr(infile, "fileno"):
        tempname = next(tempfile._get_candidate_names()) + ".mobi"
        pos = infile.tell()
        infile.seek(0)
        with open(join(tempdir, tempname), "wb") as outfile:
            shutil.copyfileobj(infile, outfile)
        infile.seek(pos)
        infile = join(tempdir, tempname)

    logger.debug("file: %s" % infile)
    fname_in = basename(infile)
    base, ext = splitext(fname_in)
    fname_out_epub = base + ".epub"
    fname_out_html = "book.html"
    fname_out_pdf = base + ".001.pdf"
    unpackBook(infile, tempdir, epubver="A")
    epub_filepath = join(tempdir, "mobi8", fname_out_epub)
    html_filepath = join(tempdir, "mobi7", fname_out_html)
    pdf_filepath = join(tempdir, fname_out_pdf)
    if exists(epub_filepath):
        return tempdir, epub_filepath
    elif exists(html_filepath):
        return tempdir, html_filepath
    elif exists(pdf_filepath):
        return tempdir, pdf_filepath
    raise ValueError("Coud not extract from %s" % infile)


if __name__ == "__main__":
    print(extract("../tests/demo.mobi"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:softtabstop=4:smarttab:expandtab

from __future__ import unicode_literals, division, absolute_import, print_function

DEBUG_USE_ORDERED_DICTIONARY = False  # OrderedDict is supoorted >= python 2.7.
""" set to True to use OrderedDict for MobiHeader.metadata."""

if DEBUG_USE_ORDERED_DICTIONARY:
    from collections import OrderedDict as dict_
else:
    dict_ = dict

from .compatibility_utils import PY2, unicode_str, hexlify, bord
from loguru import logger

if PY2:
    range = xrange

import struct
import uuid

# import the mobiunpack support libraries
from .mobi_utils import getLanguage
from .mobi_uncompress import HuffcdicReader, PalmdocReader, UncompressedReader


class unpackException(Exception):
    pass


def sortedHeaderKeys(mheader):
    hdrkeys = sorted(list(mheader.keys()), key=lambda akey: mheader[akey][0])
    return hdrkeys


# HD Containers have their own headers and their own EXTH
# this is just guesswork so far, making big assumption that
# metavalue key numbers remain the same in the CONT EXTH

# Note:  The layout of the CONT Header is still unknown
# so just deal with their EXTH sections for now


def dump_contexth(cpage, extheader):
    # determine text encoding
    codec = "windows-1252"
    codec_map = {
        1252: "windows-1252",
        65001: "utf-8",
    }
    if cpage in codec_map:
        codec = codec_map[cpage]
    if extheader == b"":
        return
    id_map_strings = {
        1: "Drm Server Id",
        2: "Drm Commerce Id",
        3: "Drm Ebookbase Book Id",
        4: "Drm Ebookbase Dep Id",
        100: "Creator",
        101: "Publisher",
        102: "Imprint",
        103: "Description",
        104: "ISBN",
        105: "Subject",
        106: "Published",
        107: "Review",
        108: "Contributor",
        109: "Rights",
        110: "SubjectCode",
        111: "Type",
        112: "Source",
        113: "ASIN",
        114: "versionNumber",
        117: "Adult",
        118: "Retail-Price",
        119: "Retail-Currency",
        120: "TSC",
        122: "fixed-layout",
        123: "book-type",
        124: "orientation-lock",
        126: "original-resolution",
        127: "zero-gutter",
        128: "zero-margin",
        129: "MetadataResourceURI",
        132: "RegionMagnification",
        150: "LendingEnabled",
        200: "DictShortName",
        501: "cdeType",
        502: "last_update_time",
        503: "Updated_Title",
        504: "CDEContentKey",
        505: "AmazonContentReference",
        506: "Title-Language",
        507: "Title-Display-Direction",
        508: "Title-Pronunciation",
        509: "Title-Collation",
        510: "Secondary-Title",
        511: "Secondary-Title-Language",
        512: "Secondary-Title-Direction",
        513: "Secondary-Title-Pronunciation",
        514: "Secondary-Title-Collation",
        515: "Author-Language",
        516: "Author-Display-Direction",
        517: "Author-Pronunciation",
        518: "Author-Collation",
        519: "Author-Type",
        520: "Publisher-Language",
        521: "Publisher-Display-Direction",
        522: "Publisher-Pronunciation",
        523: "Publisher-Collation",
        524: "Content-Language-Tag",
        525: "primary-writing-mode",
        526: "NCX-Ingested-By-Software",
        527: "page-progression-direction",
        528: "override-kindle-fonts",
        529: "Compression-Upgraded",
        530: "Soft-Hyphens-In-Content",
        531: "Dictionary_In_Langague",
        532: "Dictionary_Out_Language",
        533: "Font_Converted",
        534: "Amazon_Creator_Info",
        535: "Creator-Build-Tag",
        536: "HD-Media-Containers-Info",  # CONT_Header is 0, Ends with CONTAINER_BOUNDARY (or Asset_Type?)
        538: "Resource-Container-Fidelity",
        539: "HD-Container-Mimetype",
        540: "Sample-For_Special-Purpose",
        541: "Kindletool-Operation-Information",
        542: "Container_Id",
        543: "Asset-Type",  # FONT_CONTAINER, BW_CONTAINER, HD_CONTAINER
        544: "Unknown_544",
    }
    id_map_values = {
        115: "sample",
        116: "StartOffset",
        121: "Mobi8-Boundary-Section",
        125: "Embedded-Record-Count",
        130: "Offline-Sample",
        131: "Metadata-Record-Offset",
        201: "CoverOffset",
        202: "ThumbOffset",
        203: "HasFakeCover",
        204: "Creator-Software",
        205: "Creator-Major-Version",
        206: "Creator-Minor-Version",
        207: "Creator-Build-Number",
        401: "Clipping-Limit",
        402: "Publisher-Limit",
        404: "Text-to-Speech-Disabled",
        406: "Rental-Expiration-Time",
    }
    id_map_hexstrings = {
        208: "Watermark_(hex)",
        209: "Tamper-Proof-Keys_(hex)",
        300: "Font-Signature_(hex)",
        403: "Unknown_(403)_(hex)",
        405: "Ownership-Type_(hex)",
        407: "Unknown_(407)_(hex)",
        420: "Multimedia-Content-Reference_(hex)",
        450: "Locations_Match_(hex)",
        451: "Full-Story-Length_(hex)",
        452: "Sample-Start_Location_(hex)",
        453: "Sample-End-Location_(hex)",
    }
    _length, num_items = struct.unpack(b">LL", extheader[4:12])
    extheader = extheader[12:]
    pos = 0
    for _ in range(num_items):
        id, size = struct.unpack(b">LL", extheader[pos : pos + 8])
        content = extheader[pos + 8 : pos + size]
        if id in id_map_strings:
            name = id_map_strings[id]
            logger.debug(
                '\n    Key: "%s"\n        Value: "%s"'
                % (name, content.decode(codec, errors="replace"))
            )
        elif id in id_map_values:
            name = id_map_values[id]
            if size == 9:
                (value,) = struct.unpack(b"B", content)
                logger.debug('\n    Key: "%s"\n        Value: 0x%01x' % (name, value))
            elif size == 10:
                (value,) = struct.unpack(b">H", content)
                logger.debug('\n    Key: "%s"\n        Value: 0x%02x' % (name, value))
            elif size == 12:
                (value,) = struct.unpack(b">L", content)
                logger.debug('\n    Key: "%s"\n        Value: 0x%04x' % (name, value))
            else:
                logger.debug(
                    "\nError: Value for %s has unexpected size of %s" % (name, size)
                )
        elif id in id_map_hexstrings:
            name = id_map_hexstrings[id]
            logger.debug(
                '\n    Key: "%s"\n        Value: 0x%s' % (name, hexlify(content))
            )
        else:
            logger.debug("\nWarning: Unknown metadata with id %s found" % id)
            name = str(id) + " (hex)"
            logger.debug(
                '    Key: "%s"\n        Value: 0x%s' % (name, hexlify(content))
            )
        pos += size
    return


class MobiHeader:
    # all values are packed in big endian format
    palmdoc_header = {
        "compression_type": (0x00, b">H", 2),
        "fill0": (0x02, b">H", 2),
        "text_length": (0x04, b">L", 4),
        "text_records": (0x08, b">H", 2),
        "max_section_size": (0x0A, b">H", 2),
        "read_pos   ": (0x0C, b">L", 4),
    }

    mobi6_header = {
        "compression_type": (0x00, b">H", 2),
        "fill0": (0x02, b">H", 2),
        "text_length": (0x04, b">L", 4),
        "text_records": (0x08, b">H", 2),
        "max_section_size": (0x0A, b">H", 2),
        "crypto_type": (0x0C, b">H", 2),
        "fill1": (0x0E, b">H", 2),
        "magic": (0x10, b"4s", 4),
        "header_length (from MOBI)": (0x14, b">L", 4),
        "type": (0x18, b">L", 4),
        "codepage": (0x1C, b">L", 4),
        "unique_id": (0x20, b">L", 4),
        "version": (0x24, b">L", 4),
        "metaorthindex": (0x28, b">L", 4),
        "metainflindex": (0x2C, b">L", 4),
        "index_names": (0x30, b">L", 4),
        "index_keys": (0x34, b">L", 4),
        "extra_index0": (0x38, b">L", 4),
        "extra_index1": (0x3C, b">L", 4),
        "extra_index2": (0x40, b">L", 4),
        "extra_index3": (0x44, b">L", 4),
        "extra_index4": (0x48, b">L", 4),
        "extra_index5": (0x4C, b">L", 4),
        "first_nontext": (0x50, b">L", 4),
        "title_offset": (0x54, b">L", 4),
        "title_length": (0x58, b">L", 4),
        "language_code": (0x5C, b">L", 4),
        "dict_in_lang": (0x60, b">L", 4),
        "dict_out_lang": (0x64, b">L", 4),
        "min_version": (0x68, b">L", 4),
        "first_resc_offset": (0x6C, b">L", 4),
        "huff_offset": (0x70, b">L", 4),
        "huff_num": (0x74, b">L", 4),
        "huff_tbl_offset": (0x78, b">L", 4),
        "huff_tbl_len": (0x7C, b">L", 4),
        "exth_flags": (0x80, b">L", 4),
        "fill3_a": (0x84, b">L", 4),
        "fill3_b": (0x88, b">L", 4),
        "fill3_c": (0x8C, b">L", 4),
        "fill3_d": (0x90, b">L", 4),
        "fill3_e": (0x94, b">L", 4),
        "fill3_f": (0x98, b">L", 4),
        "fill3_g": (0x9C, b">L", 4),
        "fill3_h": (0xA0, b">L", 4),
        "unknown0": (0xA4, b">L", 4),
        "drm_offset": (0xA8, b">L", 4),
        "drm_count": (0xAC, b">L", 4),
        "drm_size": (0xB0, b">L", 4),
        "drm_flags": (0xB4, b">L", 4),
        "fill4_a": (0xB8, b">L", 4),
        "fill4_b": (0xBC, b">L", 4),
        "first_content": (0xC0, b">H", 2),
        "last_content": (0xC2, b">H", 2),
        "unknown0": (0xC4, b">L", 4),
        "fcis_offset": (0xC8, b">L", 4),
        "fcis_count": (0xCC, b">L", 4),
        "flis_offset": (0xD0, b">L", 4),
        "flis_count": (0xD4, b">L", 4),
        "unknown1": (0xD8, b">L", 4),
        "unknown2": (0xDC, b">L", 4),
        "srcs_offset": (0xE0, b">L", 4),
        "srcs_count": (0xE4, b">L", 4),
        "unknown3": (0xE8, b">L", 4),
        "unknown4": (0xEC, b">L", 4),
        "fill5": (0xF0, b">H", 2),
        "traildata_flags": (0xF2, b">H", 2),
        "ncx_index": (0xF4, b">L", 4),
        "unknown5": (0xF8, b">L", 4),
        "unknown6": (0xFC, b">L", 4),
        "datp_offset": (0x100, b">L", 4),
        "unknown7": (0x104, b">L", 4),
        "Unknown    ": (0x108, b">L", 4),
        "Unknown    ": (0x10C, b">L", 4),
        "Unknown    ": (0x110, b">L", 4),
        "Unknown    ": (0x114, b">L", 4),
        "Unknown    ": (0x118, b">L", 4),
        "Unknown    ": (0x11C, b">L", 4),
        "Unknown    ": (0x120, b">L", 4),
        "Unknown    ": (0x124, b">L", 4),
        "Unknown    ": (0x128, b">L", 4),
        "Unknown    ": (0x12C, b">L", 4),
        "Unknown    ": (0x130, b">L", 4),
        "Unknown    ": (0x134, b">L", 4),
        "Unknown    ": (0x138, b">L", 4),
        "Unknown    ": (0x11C, b">L", 4),
    }

    mobi8_header = {
        "compression_type": (0x00, b">H", 2),
        "fill0": (0x02, b">H", 2),
        "text_length": (0x04, b">L", 4),
        "text_records": (0x08, b">H", 2),
        "max_section_size": (0x0A, b">H", 2),
        "crypto_type": (0x0C, b">H", 2),
        "fill1": (0x0E, b">H", 2),
        "magic": (0x10, b"4s", 4),
        "header_length (from MOBI)": (0x14, b">L", 4),
        "type": (0x18, b">L", 4),
        "codepage": (0x1C, b">L", 4),
        "unique_id": (0x20, b">L", 4),
        "version": (0x24, b">L", 4),
        "metaorthindex": (0x28, b">L", 4),
        "metainflindex": (0x2C, b">L", 4),
        "index_names": (0x30, b">L", 4),
        "index_keys": (0x34, b">L", 4),
        "extra_index0": (0x38, b">L", 4),
        "extra_index1": (0x3C, b">L", 4),
        "extra_index2": (0x40, b">L", 4),
        "extra_index3": (0x44, b">L", 4),
        "extra_index4": (0x48, b">L", 4),
        "extra_index5": (0x4C, b">L", 4),
        "first_nontext": (0x50, b">L", 4),
        "title_offset": (0x54, b">L", 4),
        "title_length": (0x58, b">L", 4),
        "language_code": (0x5C, b">L", 4),
        "dict_in_lang": (0x60, b">L", 4),
        "dict_out_lang": (0x64, b">L", 4),
        "min_version": (0x68, b">L", 4),
        "first_resc_offset": (0x6C, b">L", 4),
        "huff_offset": (0x70, b">L", 4),
        "huff_num": (0x74, b">L", 4),
        "huff_tbl_offset": (0x78, b">L", 4),
        "huff_tbl_len": (0x7C, b">L", 4),
        "exth_flags": (0x80, b">L", 4),
        "fill3_a": (0x84, b">L", 4),
        "fill3_b": (0x88, b">L", 4),
        "fill3_c": (0x8C, b">L", 4),
        "fill3_d": (0x90, b">L", 4),
        "fill3_e": (0x94, b">L", 4),
        "fill3_f": (0x98, b">L", 4),
        "fill3_g": (0x9C, b">L", 4),
        "fill3_h": (0xA0, b">L", 4),
        "unknown0": (0xA4, b">L", 4),
        "drm_offset": (0xA8, b">L", 4),
        "drm_count": (0xAC, b">L", 4),
        "drm_size": (0xB0, b">L", 4),
        "drm_flags": (0xB4, b">L", 4),
        "fill4_a": (0xB8, b">L", 4),
        "fill4_b": (0xBC, b">L", 4),
        "fdst_offset": (0xC0, b">L", 4),
        "fdst_flow_count": (0xC4, b">L", 4),
        "fcis_offset": (0xC8, b">L", 4),
        "fcis_count": (0xCC, b">L", 4),
        "flis_offset": (0xD0, b">L", 4),
        "flis_count": (0xD4, b">L", 4),
        "unknown1": (0xD8, b">L", 4),
        "unknown2": (0xDC, b">L", 4),
        "srcs_offset": (0xE0, b">L", 4),
        "srcs_count": (0xE4, b">L", 4),
        "unknown3": (0xE8, b">L", 4),
        "unknown4": (0xEC, b">L", 4),
        "fill5": (0xF0, b">H", 2),
        "traildata_flags": (0xF2, b">H", 2),
        "ncx_index": (0xF4, b">L", 4),
        "fragment_index": (0xF8, b">L", 4),
        "skeleton_index": (0xFC, b">L", 4),
        "datp_offset": (0x100, b">L", 4),
        "guide_index": (0x104, b">L", 4),
        "Unknown    ": (0x108, b">L", 4),
        "Unknown    ": (0x10C, b">L", 4),
        "Unknown    ": (0x110, b">L", 4),
        "Unknown    ": (0x114, b">L", 4),
        "Unknown    ": (0x118, b">L", 4),
        "Unknown    ": (0x11C, b">L", 4),
        "Unknown    ": (0x120, b">L", 4),
        "Unknown    ": (0x124, b">L", 4),
        "Unknown    ": (0x128, b">L", 4),
        "Unknown    ": (0x12C, b">L", 4),
        "Unknown    ": (0x130, b">L", 4),
        "Unknown    ": (0x134, b">L", 4),
        "Unknown    ": (0x138, b">L", 4),
        "Unknown    ": (0x11C, b">L", 4),
    }

    palmdoc_header_sorted_keys = sortedHeaderKeys(palmdoc_header)
    mobi6_header_sorted_keys = sortedHeaderKeys(mobi6_header)
    mobi8_header_sorted_keys = sortedHeaderKeys(mobi8_header)

    id_map_strings = {
        1: "Drm Server Id",
        2: "Drm Commerce Id",
        3: "Drm Ebookbase Book Id",
        4: "Drm Ebookbase Dep Id",
        100: "Creator",
        101: "Publisher",
        102: "Imprint",
        103: "Description",
        104: "ISBN",
        105: "Subject",
        106: "Published",
        107: "Review",
        108: "Contributor",
        109: "Rights",
        110: "SubjectCode",
        111: "Type",
        112: "Source",
        113: "ASIN",
        114: "versionNumber",
        117: "Adult",
        118: "Retail-Price",
        119: "Retail-Currency",
        120: "TSC",
        122: "fixed-layout",
        123: "book-type",
        124: "orientation-lock",
        126: "original-resolution",
        127: "zero-gutter",
        128: "zero-margin",
        129: "MetadataResourceURI",
        132: "RegionMagnification",
        150: "LendingEnabled",
        200: "DictShortName",
        501: "cdeType",
        502: "last_update_time",
        503: "Updated_Title",
        504: "CDEContentKey",
        505: "AmazonContentReference",
        506: "Title-Language",
        507: "Title-Display-Direction",
        508: "Title-Pronunciation",
        509: "Title-Collation",
        510: "Secondary-Title",
        511: "Secondary-Title-Language",
        512: "Secondary-Title-Direction",
        513: "Secondary-Title-Pronunciation",
        514: "Secondary-Title-Collation",
        515: "Author-Language",
        516: "Author-Display-Direction",
        517: "Author-Pronunciation",
        518: "Author-Collation",
        519: "Author-Type",
        520: "Publisher-Language",
        521: "Publisher-Display-Direction",
        522: "Publisher-Pronunciation",
        523: "Publisher-Collation",
        524: "Content-Language-Tag",
        525: "primary-writing-mode",
        526: "NCX-Ingested-By-Software",
        527: "page-progression-direction",
        528: "override-kindle-fonts",
        529: "Compression-Upgraded",
        530: "Soft-Hyphens-In-Content",
        531: "Dictionary_In_Langague",
        532: "Dictionary_Out_Language",
        533: "Font_Converted",
        534: "Amazon_Creator_Info",
        535: "Creator-Build-Tag",
        536: "HD-Media-Containers-Info",  # CONT_Header is 0, Ends with CONTAINER_BOUNDARY (or Asset_Type?)
        538: "Resource-Container-Fidelity",
        539: "HD-Container-Mimetype",
        540: "Sample-For_Special-Purpose",
        541: "Kindletool-Operation-Information",
        542: "Container_Id",
        543: "Asset-Type",  # FONT_CONTAINER, BW_CONTAINER, HD_CONTAINER
        544: "Unknown_544",
    }
    id_map_values = {
        115: "sample",
        116: "StartOffset",
        121: "Mobi8-Boundary-Section",
        125: "Embedded-Record-Count",
        130: "Offline-Sample",
        131: "Metadata-Record-Offset",
        201: "CoverOffset",
        202: "ThumbOffset",
        203: "HasFakeCover",
        204: "Creator-Software",
        205: "Creator-Major-Version",
        206: "Creator-Minor-Version",
        207: "Creator-Build-Number",
        401: "Clipping-Limit",
        402: "Publisher-Limit",
        404: "Text-to-Speech-Disabled",
        406: "Rental-Expiration-Time",
    }
    id_map_hexstrings = {
        208: "Watermark_(hex)",
        209: "Tamper-Proof-Keys_(hex)",
        300: "Font-Signature_(hex)",
        403: "Unknown_(403)_(hex)",
        405: "Ownership-Type_(hex)",
        407: "Unknown_(407)_(hex)",
        420: "Multimedia-Content-Reference_(hex)",
        450: "Locations_Match_(hex)",
        451: "Full-Story-Length_(hex)",
        452: "Sample-Start_Location_(hex)",
        453: "Sample-End-Location_(hex)",
    }

    def __init__(self, sect, sectNumber):
        self.sect = sect
        self.start = sectNumber
        self.header = self.sect.loadSection(self.start)
        if len(self.header) > 20 and self.header[16:20] == b"MOBI":
            self.sect.setsectiondescription(0, "Mobipocket Header")
            self.palm = False
        elif self.sect.ident == b"TEXtREAd":
            self.sect.setsectiondescription(0, "PalmDOC Header")
            self.palm = True
        else:
            raise unpackException("Unknown File Format")

        (self.records,) = struct.unpack_from(b">H", self.header, 0x8)

        # set defaults in case this is a PalmDOC
        self.title = self.sect.palmname.decode("latin-1", errors="replace")
        self.length = len(self.header) - 16
        self.type = 3
        self.codepage = 1252
        self.codec = "windows-1252"
        self.unique_id = 0
        self.version = 0
        self.hasExth = False
        self.exth = b""
        self.exth_offset = self.length + 16
        self.exth_length = 0
        self.crypto_type = 0
        self.firstnontext = self.start + self.records + 1
        self.firstresource = self.start + self.records + 1
        self.ncxidx = 0xFFFFFFFF
        self.metaOrthIndex = 0xFFFFFFFF
        self.metaInflIndex = 0xFFFFFFFF
        self.skelidx = 0xFFFFFFFF
        self.fragidx = 0xFFFFFFFF
        self.guideidx = 0xFFFFFFFF
        self.fdst = 0xFFFFFFFF
        self.mlstart = self.sect.loadSection(self.start + 1)[:4]
        self.rawSize = 0
        self.metadata = dict_()

        # set up for decompression/unpacking
        (self.compression,) = struct.unpack_from(b">H", self.header, 0x0)
        if self.compression == 0x4448:
            reader = HuffcdicReader()
            huffoff, huffnum = struct.unpack_from(b">LL", self.header, 0x70)
            huffoff = huffoff + self.start
            self.sect.setsectiondescription(huffoff, "Huffman Compression Seed")
            reader.loadHuff(self.sect.loadSection(huffoff))
            for i in range(1, huffnum):
                self.sect.setsectiondescription(
                    huffoff + i, "Huffman CDIC Compression Seed %d" % i
                )
                reader.loadCdic(self.sect.loadSection(huffoff + i))
            self.unpack = reader.unpack
        elif self.compression == 2:
            self.unpack = PalmdocReader().unpack
        elif self.compression == 1:
            self.unpack = UncompressedReader().unpack
        else:
            raise unpackException("invalid compression type: 0x%4x" % self.compression)

        if self.palm:
            return

        (
            self.length,
            self.type,
            self.codepage,
            self.unique_id,
            self.version,
        ) = struct.unpack(b">LLLLL", self.header[20:40])
        codec_map = {
            1252: "windows-1252",
            65001: "utf-8",
        }
        if self.codepage in codec_map:
            self.codec = codec_map[self.codepage]

        # title
        toff, tlen = struct.unpack(b">II", self.header[0x54:0x5C])
        tend = toff + tlen
        self.title = self.header[toff:tend].decode(self.codec, errors="replace")

        (exth_flag,) = struct.unpack(b">L", self.header[0x80:0x84])
        self.hasExth = exth_flag & 0x40
        self.exth_offset = self.length + 16
        self.exth_length = 0
        if self.hasExth:
            (self.exth_length,) = struct.unpack_from(
                b">L", self.header, self.exth_offset + 4
            )
            self.exth_length = (
                (self.exth_length + 3) >> 2
            ) << 2  # round to next 4 byte boundary
            self.exth = self.header[
                self.exth_offset : self.exth_offset + self.exth_length
            ]

        # parse the exth / metadata
        self.parseMetaData()

        # self.mlstart = self.sect.loadSection(self.start+1)
        # self.mlstart = self.mlstart[0:4]
        (self.crypto_type,) = struct.unpack_from(b">H", self.header, 0xC)

        # Start sector for additional files such as images, fonts, resources, etc
        # Can be missing so fall back to default set previously
        (ofst,) = struct.unpack_from(b">L", self.header, 0x6C)
        if ofst != 0xFFFFFFFF:
            self.firstresource = ofst + self.start
        (ofst,) = struct.unpack_from(b">L", self.header, 0x50)
        if ofst != 0xFFFFFFFF:
            self.firstnontext = ofst + self.start

        if self.isPrintReplica():
            return

        if self.version < 8:
            # Dictionary metaOrthIndex
            (self.metaOrthIndex,) = struct.unpack_from(b">L", self.header, 0x28)
            if self.metaOrthIndex != 0xFFFFFFFF:
                self.metaOrthIndex += self.start

            # Dictionary metaInflIndex
            (self.metaInflIndex,) = struct.unpack_from(b">L", self.header, 0x2C)
            if self.metaInflIndex != 0xFFFFFFFF:
                self.metaInflIndex += self.start

        # handle older headers without any ncxindex info and later
        # specifically 0xe4 headers
        if self.length + 16 < 0xF8:
            return

        # NCX Index
        (self.ncxidx,) = struct.unpack(b">L", self.header[0xF4:0xF8])
        if self.ncxidx != 0xFFFFFFFF:
            self.ncxidx += self.start

        # K8 specific Indexes
        if self.start != 0 or self.version == 8:
            # Index into <xml> file skeletons in RawML
            (self.skelidx,) = struct.unpack_from(b">L", self.header, 0xFC)
            if self.skelidx != 0xFFFFFFFF:
                self.skelidx += self.start

            # Index into <div> sections in RawML
            (self.fragidx,) = struct.unpack_from(b">L", self.header, 0xF8)
            if self.fragidx != 0xFFFFFFFF:
                self.fragidx += self.start

            # Index into Other files
            (self.guideidx,) = struct.unpack_from(b">L", self.header, 0x104)
            if self.guideidx != 0xFFFFFFFF:
                self.guideidx += self.start

            # dictionaries do not seem to use the same approach in K8's
            # so disable them
            self.metaOrthIndex = 0xFFFFFFFF
            self.metaInflIndex = 0xFFFFFFFF

            # need to use the FDST record to find out how to properly unpack
            # the rawML into pieces
            # it is simply a table of start and end locations for each flow piece
            (self.fdst,) = struct.unpack_from(b">L", self.header, 0xC0)
            (self.fdstcnt,) = struct.unpack_from(b">L", self.header, 0xC4)
            # if cnt is 1 or less, fdst section mumber can be garbage
            if self.fdstcnt <= 1:
                self.fdst = 0xFFFFFFFF
            if self.fdst != 0xFFFFFFFF:
                self.fdst += self.start
                # setting of fdst section description properly handled in mobi_kf8proc

    def dump_exth(self):
        # determine text encoding
        codec = self.codec
        if (not self.hasExth) or (self.exth_length) == 0 or (self.exth == b""):
            return
        (num_items,) = struct.unpack(b">L", self.exth[8:12])
        pos = 12
        logger.debug("Key Size Decription                     Value")
        for _ in range(num_items):
            id, size = struct.unpack(b">LL", self.exth[pos : pos + 8])
            contentsize = size - 8
            content = self.exth[pos + 8 : pos + size]
            if id in MobiHeader.id_map_strings:
                exth_name = MobiHeader.id_map_strings[id]
                logger.debug(
                    "{0: >3d} {1: >4d} {2: <30s} {3:s}".format(
                        id,
                        contentsize,
                        exth_name,
                        content.decode(codec, errors="replace"),
                    )
                )
            elif id in MobiHeader.id_map_values:
                exth_name = MobiHeader.id_map_values[id]
                if size == 9:
                    (value,) = struct.unpack(b"B", content)
                    logger.debug(
                        "{0:3d} byte {1:<30s} {2:d}".format(id, exth_name, value)
                    )
                elif size == 10:
                    (value,) = struct.unpack(b">H", content)
                    logger.debug(
                        "{0:3d} word {1:<30s} 0x{2:0>4X} ({2:d})".format(
                            id, exth_name, value
                        )
                    )
                elif size == 12:
                    (value,) = struct.unpack(b">L", content)
                    logger.debug(
                        "{0:3d} long {1:<30s} 0x{2:0>8X} ({2:d})".format(
                            id, exth_name, value
                        )
                    )
                else:
                    logger.debug(
                        "{0: >3d} {1: >4d} {2: <30s} (0x{3:s})".format(
                            id,
                            contentsize,
                            "Bad size for " + exth_name,
                            hexlify(content),
                        )
                    )
            elif id in MobiHeader.id_map_hexstrings:
                exth_name = MobiHeader.id_map_hexstrings[id]
                logger.debug(
                    "{0:3d} {1:4d} {2:<30s} 0x{3:s}".format(
                        id, contentsize, exth_name, hexlify(content)
                    )
                )
            else:
                exth_name = "Unknown EXTH ID {0:d}".format(id)
                logger.debug(
                    "{0: >3d} {1: >4d} {2: <30s} 0x{3:s}".format(
                        id, contentsize, exth_name, hexlify(content)
                    )
                )
            pos += size
        return

    def dumpheader(self):
        # first 16 bytes are not part of the official mobiheader
        # but we will treat it as such
        # so section 0 is 16 (decimal) + self.length in total == at least 0x108 bytes for Mobi 8 headers
        logger.debug(
            "Dumping section %d, Mobipocket Header version: %d, total length %d"
            % (self.start, self.version, self.length + 16)
        )
        self.hdr = {}
        # set it up for the proper header version
        if self.version == 0:
            self.mobi_header = MobiHeader.palmdoc_header
            self.mobi_header_sorted_keys = MobiHeader.palmdoc_header_sorted_keys
        elif self.version < 8:
            self.mobi_header = MobiHeader.mobi6_header
            self.mobi_header_sorted_keys = MobiHeader.mobi6_header_sorted_keys
        else:
            self.mobi_header = MobiHeader.mobi8_header
            self.mobi_header_sorted_keys = MobiHeader.mobi8_header_sorted_keys

        # parse the header information
        for key in self.mobi_header_sorted_keys:
            (pos, format, tot_len) = self.mobi_header[key]
            if pos < (self.length + 16):
                (val,) = struct.unpack_from(format, self.header, pos)
                self.hdr[key] = val

        if "title_offset" in self.hdr:
            title_offset = self.hdr["title_offset"]
            title_length = self.hdr["title_length"]
        else:
            title_offset = 0
            title_length = 0
        if title_offset == 0:
            title_offset = len(self.header)
            title_length = 0
            self.title = self.sect.palmname.decode("latin-1", errors="replace")
        else:
            self.title = self.header[title_offset : title_offset + title_length].decode(
                self.codec, errors="replace"
            )
            # title record always padded with two nul bytes and then padded with nuls to next 4 byte boundary
            title_length = ((title_length + 2 + 3) >> 2) << 2

        self.extra1 = self.header[self.exth_offset + self.exth_length : title_offset]
        self.extra2 = self.header[title_offset + title_length :]

        logger.debug("Mobipocket header from section %d" % self.start)
        logger.debug("     Offset  Value Hex Dec        Description")
        for key in self.mobi_header_sorted_keys:
            (pos, format, tot_len) = self.mobi_header[key]
            if pos < (self.length + 16):
                if key != "magic":
                    fmt_string = (
                        "0x{0:0>3X} ({0:3d}){1: >"
                        + str(9 - 2 * tot_len)
                        + "s}0x{2:0>"
                        + str(2 * tot_len)
                        + "X} {2:10d} {3:s}"
                    )
                else:
                    self.hdr[key] = unicode_str(self.hdr[key])
                    fmt_string = "0x{0:0>3X} ({0:3d}){2:>11s}            {3:s}"
                logger.debug(fmt_string.format(pos, " ", self.hdr[key], key))
        logger.debug("")

        if self.exth_length > 0:
            logger.debug(
                "EXTH metadata, offset %d, padded length %d"
                % (self.exth_offset, self.exth_length)
            )
            self.dump_exth()
            logger.debug("")

        if len(self.extra1) > 0:
            logger.debug(
                "Extra data between EXTH and Title, length %d" % len(self.extra1)
            )
            logger.debug(hexlify(self.extra1))
            logger.debug("")

        if title_length > 0:
            logger.debug(
                "Title in header at offset %d, padded length %d: '%s'"
                % (title_offset, title_length, self.title)
            )
            logger.debug("")

        if len(self.extra2) > 0:
            print(
                "Extra data between Title and end of header, length %d"
                % len(self.extra2)
            )
            print(hexlify(self.extra2))
            print("")

    def isPrintReplica(self):
        return self.mlstart[0:4] == b"%MOP"

    def isK8(self):
        return self.start != 0 or self.version == 8

    def isEncrypted(self):
        return self.crypto_type != 0

    def hasNCX(self):
        return self.ncxidx != 0xFFFFFFFF

    def isDictionary(self):
        return self.metaOrthIndex != 0xFFFFFFFF

    def getncxIndex(self):
        return self.ncxidx

    def decompress(self, data):
        return self.unpack(data)

    def Language(self):
        langcode = struct.unpack(b"!L", self.header[0x5C:0x60])[0]
        langid = langcode & 0xFF
        sublangid = (langcode >> 8) & 0xFF
        return getLanguage(langid, sublangid)

    def DictInLanguage(self):
        if self.isDictionary():
            langcode = struct.unpack(b"!L", self.header[0x60:0x64])[0]
            langid = langcode & 0xFF
            sublangid = (langcode >> 10) & 0xFF
            if langid != 0:
                return getLanguage(langid, sublangid)
        return False

    def DictOutLanguage(self):
        if self.isDictionary():
            langcode = struct.unpack(b"!L", self.header[0x64:0x68])[0]
            langid = langcode & 0xFF
            sublangid = (langcode >> 10) & 0xFF
            if langid != 0:
                return getLanguage(langid, sublangid)
        return False

    def getRawML(self):
        def getSizeOfTrailingDataEntry(data):
            num = 0
            for v in data[-4:]:
                if bord(v) & 0x80:
                    num = 0
                num = (num << 7) | (bord(v) & 0x7F)
            return num

        def trimTrailingDataEntries(data):
            for _ in range(trailers):
                num = getSizeOfTrailingDataEntry(data)
                data = data[:-num]
            if multibyte:
                num = (ord(data[-1:]) & 3) + 1
                data = data[:-num]
            return data

        multibyte = 0
        trailers = 0
        if self.sect.ident == b"BOOKMOBI":
            (mobi_length,) = struct.unpack_from(b">L", self.header, 0x14)
            (mobi_version,) = struct.unpack_from(b">L", self.header, 0x68)
            if (mobi_length >= 0xE4) and (mobi_version >= 5):
                (flags,) = struct.unpack_from(b">H", self.header, 0xF2)
                multibyte = flags & 1
                while flags > 1:
                    if flags & 2:
                        trailers += 1
                    flags = flags >> 1
        # get raw mobi markup languge
        logger.debug("Unpacking raw markup language")
        dataList = []
        # offset = 0
        for i in range(1, self.records + 1):
            data = trimTrailingDataEntries(self.sect.loadSection(self.start + i))
            dataList.append(self.unpack(data))
            if self.isK8():
                self.sect.setsectiondescription(
                    self.start + i, "KF8 Text Section {0:d}".format(i)
                )
            elif self.version == 0:
                self.sect.setsectiondescription(
                    self.start + i, "PalmDOC Text Section {0:d}".format(i)
                )
            else:
                self.sect.setsectiondescription(
                    self.start + i, "Mobipocket Text Section {0:d}".format(i)
                )
        rawML = b"".join(dataList)
        self.rawSize = len(rawML)
        return rawML

    # all metadata is stored in a dictionary with key and returns a *list* of values
    # a list is used to allow for multiple creators, multiple contributors, etc
    def parseMetaData(self):
        def addValue(name, value):
            if name not in self.metadata:
                self.metadata[name] = [value]
            else:
                self.metadata[name].append(value)

        codec = self.codec
        if self.hasExth:
            extheader = self.exth
            _length, num_items = struct.unpack(b">LL", extheader[4:12])
            extheader = extheader[12:]
            pos = 0
            for _ in range(num_items):
                id, size = struct.unpack(b">LL", extheader[pos : pos + 8])
                content = extheader[pos + 8 : pos + size]
                if id in MobiHeader.id_map_strings:
                    name = MobiHeader.id_map_strings[id]
                    addValue(name, content.decode(codec, errors="replace"))
                elif id in MobiHeader.id_map_values:
                    name = MobiHeader.id_map_values[id]
                    if size == 9:
                        (value,) = struct.unpack(b"B", content)
                        addValue(name, unicode_str(str(value)))
                    elif size == 10:
                        (value,) = struct.unpack(b">H", content)
                        addValue(name, unicode_str(str(value)))
                    elif size == 12:
                        (value,) = struct.unpack(b">L", content)
                        # handle special case of missing CoverOffset or missing ThumbOffset
                        if id == 201 or id == 202:
                            if value != 0xFFFFFFFF:
                                addValue(name, unicode_str(str(value)))
                        else:
                            addValue(name, unicode_str(str(value)))
                    else:
                        print(
                            "Warning: Bad key, size, value combination detected in EXTH ",
                            id,
                            size,
                            hexlify(content),
                        )
                        addValue(name, hexlify(content))
                elif id in MobiHeader.id_map_hexstrings:
                    name = MobiHeader.id_map_hexstrings[id]
                    addValue(name, hexlify(content))
                else:
                    name = unicode_str(str(id)) + " (hex)"
                    addValue(name, hexlify(content))
                pos += size

        # add the basics to the metadata each as a list element
        self.metadata["Language"] = [self.Language()]
        self.metadata["Title"] = [unicode_str(self.title, self.codec)]
        self.metadata["Codec"] = [self.codec]
        self.metadata["UniqueID"] = [unicode_str(str(self.unique_id))]
        # if no asin create one using a uuid
        if "ASIN" not in self.metadata:
            self.metadata["ASIN"] = [unicode_str(str(uuid.uuid4()))]
        # if no cdeType set it to "EBOK"
        if "cdeType" not in self.metadata:
            self.metadata["cdeType"] = ["EBOK"]

    def getMetaData(self):
        return self.metadata

    def describeHeader(self, DUMP):
        logger.debug("Mobi Version: %s" % self.version)
        logger.debug("Codec: %s" % self.codec)
        logger.debug("Title: %s" % self.title)
        if "Updated_Title" in self.metadata:
            logger.debug("EXTH Title: %s" % self.metadata["Updated_Title"][0])
        if self.compression == 0x4448:
            logger.debug("Huffdic compression")
        elif self.compression == 2:
            logger.debug("Palmdoc compression")
        elif self.compression == 1:
            logger.debug("No compression")
        if DUMP:
            self.dumpheader()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:softtabstop=4:smarttab:expandtab

from __future__ import unicode_literals, division, absolute_import, print_function

from .compatibility_utils import text_type

from . import unipath
from .unipath import pathof

DUMP = False
""" Set to True to dump all possible information. """

import os

import re

# note: re requites the pattern to be the exact same type as the data to be searched in python3
# but u"" is not allowed for the pattern itself only b""

import zipfile
import binascii
from .mobi_utils import mangle_fonts


class unpackException(Exception):
    pass


class ZipInfo(zipfile.ZipInfo):
    def __init__(self, *args, **kwargs):
        if "compress_type" in kwargs:
            compress_type = kwargs.pop("compress_type")
        super(ZipInfo, self).__init__(*args, **kwargs)
        self.compress_type = compress_type


class fileNames:
    def __init__(self, infile, outdir):
        self.infile = infile
        self.outdir = outdir
        if not unipath.exists(self.outdir):
            unipath.mkdir(self.outdir)
        self.mobi7dir = os.path.join(self.outdir, "mobi7")
        if not unipath.exists(self.mobi7dir):
            unipath.mkdir(self.mobi7dir)
        self.imgdir = os.path.join(self.mobi7dir, "Images")
        if not unipath.exists(self.imgdir):
            unipath.mkdir(self.imgdir)
        self.hdimgdir = os.path.join(self.outdir, "HDImages")
        if not unipath.exists(self.hdimgdir):
            unipath.mkdir(self.hdimgdir)
        self.outbase = os.path.join(
            self.outdir, os.path.splitext(os.path.split(infile)[1])[0]
        )

    def getInputFileBasename(self):
        return os.path.splitext(os.path.basename(self.infile))[0]

    def makeK8Struct(self):
        self.k8dir = os.path.join(self.outdir, "mobi8")
        if not unipath.exists(self.k8dir):
            unipath.mkdir(self.k8dir)
        self.k8metainf = os.path.join(self.k8dir, "META-INF")
        if not unipath.exists(self.k8metainf):
            unipath.mkdir(self.k8metainf)
        self.k8oebps = os.path.join(self.k8dir, "OEBPS")
        if not unipath.exists(self.k8oebps):
            unipath.mkdir(self.k8oebps)
        self.k8images = os.path.join(self.k8oebps, "Images")
        if not unipath.exists(self.k8images):
            unipath.mkdir(self.k8images)
        self.k8fonts = os.path.join(self.k8oebps, "Fonts")
        if not unipath.exists(self.k8fonts):
            unipath.mkdir(self.k8fonts)
        self.k8styles = os.path.join(self.k8oebps, "Styles")
        if not unipath.exists(self.k8styles):
            unipath.mkdir(self.k8styles)
        self.k8text = os.path.join(self.k8oebps, "Text")
        if not unipath.exists(self.k8text):
            unipath.mkdir(self.k8text)

    # recursive zip creation support routine
    def zipUpDir(self, myzip, tdir, localname):
        currentdir = tdir
        if localname != "":
            currentdir = os.path.join(currentdir, localname)
        list = unipath.listdir(currentdir)
        for file in list:
            afilename = file
            localfilePath = os.path.join(localname, afilename)
            realfilePath = os.path.join(currentdir, file)
            if unipath.isfile(realfilePath):
                myzip.write(
                    pathof(realfilePath), pathof(localfilePath), zipfile.ZIP_DEFLATED
                )
            elif unipath.isdir(realfilePath):
                self.zipUpDir(myzip, tdir, localfilePath)

    def makeEPUB(self, usedmap, obfuscate_data, uid):
        bname = os.path.join(self.k8dir, self.getInputFileBasename() + ".epub")
        # Create an encryption key for Adobe font obfuscation
        # based on the epub's uid
        if isinstance(uid, text_type):
            uid = uid.encode("ascii")
        if obfuscate_data:
            key = re.sub(br"[^a-fA-F0-9]", b"", uid)
            key = binascii.unhexlify((key + key)[:32])

        # copy over all images and fonts that are actually used in the ebook
        # and remove all font files from mobi7 since not supported
        imgnames = unipath.listdir(self.imgdir)
        for name in imgnames:
            if usedmap.get(name, "not used") == "used":
                filein = os.path.join(self.imgdir, name)
                if name.endswith(".ttf"):
                    fileout = os.path.join(self.k8fonts, name)
                elif name.endswith(".otf"):
                    fileout = os.path.join(self.k8fonts, name)
                elif name.endswith(".failed"):
                    fileout = os.path.join(self.k8fonts, name)
                else:
                    fileout = os.path.join(self.k8images, name)
                data = b""
                with open(pathof(filein), "rb") as f:
                    data = f.read()
                if obfuscate_data:
                    if name in obfuscate_data:
                        data = mangle_fonts(key, data)
                open(pathof(fileout), "wb").write(data)
                if name.endswith(".ttf") or name.endswith(".otf"):
                    os.remove(pathof(filein))

        # opf file name hard coded to "content.opf"
        container = '<?xml version="1.0" encoding="UTF-8"?>\n'
        container += '<container version="1.0" xmlns="urn:oasis:names:tc:opendocument:xmlns:container">\n'
        container += "    <rootfiles>\n"
        container += '<rootfile full-path="OEBPS/content.opf" media-type="application/oebps-package+xml"/>'
        container += "    </rootfiles>\n</container>\n"
        fileout = os.path.join(self.k8metainf, "container.xml")
        with open(pathof(fileout), "wb") as f:
            f.write(container.encode("utf-8"))

        if obfuscate_data:
            encryption = '<encryption xmlns="urn:oasis:names:tc:opendocument:xmlns:container" \
xmlns:enc="http://www.w3.org/2001/04/xmlenc#" xmlns:deenc="http://ns.adobe.com/digitaleditions/enc">\n'
            for font in obfuscate_data:
                encryption += "  <enc:EncryptedData>\n"
                encryption += '    <enc:EncryptionMethod Algorithm="http://ns.adobe.com/pdf/enc#RC"/>\n'
                encryption += "    <enc:CipherData>\n"
                encryption += (
                    '      <enc:CipherReference URI="OEBPS/Fonts/' + font + '"/>\n'
                )
                encryption += "    </enc:CipherData>\n"
                encryption += "  </enc:EncryptedData>\n"
            encryption += "</encryption>\n"
            fileout = os.path.join(self.k8metainf, "encryption.xml")
            with open(pathof(fileout), "wb") as f:
                f.write(encryption.encode("utf-8"))

        # ready to build epub
        self.outzip = zipfile.ZipFile(pathof(bname), "w")

        # add the mimetype file uncompressed
        mimetype = b"application/epub+zip"
        fileout = os.path.join(self.k8dir, "mimetype")
        with open(pathof(fileout), "wb") as f:
            f.write(mimetype)
        nzinfo = ZipInfo("mimetype", compress_type=zipfile.ZIP_STORED)
        nzinfo.external_attr = 0o600 << 16  # make this a normal file
        self.outzip.writestr(nzinfo, mimetype)
        self.zipUpDir(self.outzip, self.k8dir, "META-INF")
        self.zipUpDir(self.outzip, self.k8dir, "OEBPS")
        self.outzip.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:softtabstop=4:smarttab:expandtab

from __future__ import unicode_literals, division, absolute_import, print_function

from .compatibility_utils import PY2, bchr, lmap, bstr

if PY2:
    range = xrange

import struct

# note:  struct pack, unpack, unpack_from all require bytestring format
# data all the way up to at least python 2.7.5, python 3 okay with bytestring


class unpackException(Exception):
    pass


class UncompressedReader:
    def unpack(self, data):
        return data


class PalmdocReader:
    def unpack(self, i):
        o, p = b"", 0
        while p < len(i):
            # for python 3 must use slice since i[p] returns int while slice returns character
            c = ord(i[p : p + 1])
            p += 1
            if c >= 1 and c <= 8:
                o += i[p : p + c]
                p += c
            elif c < 128:
                o += bchr(c)
            elif c >= 192:
                o += b" " + bchr(c ^ 128)
            else:
                if p < len(i):
                    c = (c << 8) | ord(i[p : p + 1])
                    p += 1
                    m = (c >> 3) & 0x07FF
                    n = (c & 7) + 3
                    if m > n:
                        o += o[-m : n - m]
                    else:
                        for _ in range(n):
                            # because of completely ass-backwards decision by python mainters for python 3
                            # we must use slice for bytes as i[p] returns int while slice returns character
                            if m == 1:
                                o += o[-m:]
                            else:
                                o += o[-m : -m + 1]
        return o


class HuffcdicReader:
    q = struct.Struct(b">Q").unpack_from

    def loadHuff(self, huff):
        if huff[0:8] != b"HUFF\x00\x00\x00\x18":
            raise unpackException("invalid huff header")
        off1, off2 = struct.unpack_from(b">LL", huff, 8)

        def dict1_unpack(v):
            codelen, term, maxcode = v & 0x1F, v & 0x80, v >> 8
            assert codelen != 0
            if codelen <= 8:
                assert term
            maxcode = ((maxcode + 1) << (32 - codelen)) - 1
            return (codelen, term, maxcode)

        self.dict1 = lmap(dict1_unpack, struct.unpack_from(b">256L", huff, off1))

        dict2 = struct.unpack_from(b">64L", huff, off2)
        self.mincode, self.maxcode = (), ()
        for codelen, mincode in enumerate((0,) + dict2[0::2]):
            self.mincode += (mincode << (32 - codelen),)
        for codelen, maxcode in enumerate((0,) + dict2[1::2]):
            self.maxcode += (((maxcode + 1) << (32 - codelen)) - 1,)

        self.dictionary = []

    def loadCdic(self, cdic):
        if cdic[0:8] != b"CDIC\x00\x00\x00\x10":
            raise unpackException("invalid cdic header")
        phrases, bits = struct.unpack_from(b">LL", cdic, 8)
        n = min(1 << bits, phrases - len(self.dictionary))
        h = struct.Struct(b">H").unpack_from

        def getslice(off):
            (blen,) = h(cdic, 16 + off)
            slice = cdic[18 + off : 18 + off + (blen & 0x7FFF)]
            return (slice, blen & 0x8000)

        self.dictionary += lmap(
            getslice, struct.unpack_from(bstr(">%dH" % n), cdic, 16)
        )

    def unpack(self, data):
        q = HuffcdicReader.q

        bitsleft = len(data) * 8
        data += b"\x00\x00\x00\x00\x00\x00\x00\x00"
        pos = 0
        (x,) = q(data, pos)
        n = 32

        s = b""
        while True:
            if n <= 0:
                pos += 4
                (x,) = q(data, pos)
                n += 32
            code = (x >> n) & ((1 << 32) - 1)

            codelen, term, maxcode = self.dict1[code >> 24]
            if not term:
                while code < self.mincode[codelen]:
                    codelen += 1
                maxcode = self.maxcode[codelen]

            n -= codelen
            bitsleft -= codelen
            if bitsleft < 0:
                break

            r = (maxcode - code) >> (32 - codelen)
            slice, flag = self.dictionary[r]
            if not flag:
                self.dictionary[r] = None
                slice = self.unpack(slice)
                self.dictionary[r] = (slice, 1)
            s += slice
        return s



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mobi']

package_data = \
{'': ['*']}

install_requires = \
['loguru>=0.4,<0.5']

entry_points = \
{'console_scripts': ['mobiunpack = mobi.kindleunpack:main']}

setup_kwargs = {
    'name': 'mobi',
    'version': '0.3.0',
    'description': 'unpack unencrypted mobi files',
    'long_description': '# mobi - library for unpacking unencrypted mobi files\n\n[![Version](https://img.shields.io/pypi/v/mobi.svg)](https://pypi.python.org/pypi/mobi/)\n[![Downloads](https://pepy.tech/badge/mobi)](https://pepy.tech/project/mobi)\n\n> A fork of [KindleUnpack](https://github.com/kevinhendricks/KindleUnpack) which removes the GUI part and makes it available as a python library via [PyPi](https://pypi.org/project/mobi/) for easy unpacking of mobi files.\n\n## Usage\n\n### As library\n\n```python\nimport mobi\n\ntempdir, filepath = mobi.extract("mybook.mobi")\n```\n\n\'tempdir\' is the path where the mobi is unpacked\n\'filepath\' is the path to either an epub, html or pdf file depending on the mobi type \n\n| NOTE: You are responsible to delete the generated tempdir! |\n| --- |\n\n### From the command line\n\nThe installer also creates a console script entrypoint that wraps the original KindleUnpack\n\n```console\n$ mobiunpack\nKindleUnpack v0.82\n   Based on initial mobipocket version Copyright © 2009 Charles M. Hannum <root@ihack.net>\n   Extensive Extensions and Improvements Copyright © 2009-2014\n       by:  P. Durrant, K. Hendricks, S. Siebert, fandrieu, DiapDealer, nickredding, tkeo.\n   This program is free software: you can redistribute it and/or modify\n   it under the terms of the GNU General Public License as published by\n   the Free Software Foundation, version 3.\n\nDescription:\n  Unpacks an unencrypted Kindle/MobiPocket ebook to html and images\n  or an unencrypted Kindle/Print Replica ebook to PDF and images\n  into the specified output folder.\nUsage:\n  mobiunpack -r -s -p apnxfile -d -h --epub_version= infile [outdir]\nOptions:\n    -h                 print this help message\n    -i                 use HD Images, if present, to overwrite reduced resolution images\n    -s                 split combination mobis into mobi7 and mobi8 ebooks\n    -p APNXFILE        path to an .apnx file associated with the azw3 input (optional)\n    --epub_version=    specify epub version to unpack to: 2, 3, A (for automatic) or\n                         F (force to fit to epub2 definitions), default is 2\n    -d                 dump headers and other info to output and extra files\n    -r                 write raw data to the output folder\n```\n\n### [0.3.0] - 2020-03-02\n\n- Add support for mobi7 only files\n- Add experimental support for mobi print replica files\n- Add support for file-like objects\n\n\n### [0.2.0] - 2020-03-02\n\n- Minimal working \'extract\' function and \'mobiunpack\' console wrapper\n- Replace most print calls with logging\n\n### [0.1.0] - 2020-03-02\n\n- Empty package registered on pypi\n\n## License\n\nGPL-3.0-only\n\nAll credits for the hard work go to https://github.com/kevinhendricks/KindleUnpack\n',
    'author': 'Titusz Pan',
    'author_email': 'tp@py7.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github/iscc/mobi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:softtabstop=4:smarttab:expandtab

from __future__ import unicode_literals, division, absolute_import, print_function

DEBUG_USE_ORDERED_DICTIONARY = False  # OrderedDict is supoorted >= python 2.7.
""" set to True to use OrderedDict for K8RESCProcessor.parsetag.tattr."""

if DEBUG_USE_ORDERED_DICTIONARY:
    from collections import OrderedDict as dict_
else:
    dict_ = dict

from .compatibility_utils import unicode_str
from loguru import logger

from .mobi_utils import fromBase32

_OPF_PARENT_TAGS = [
    "xml",
    "package",
    "metadata",
    "dc-metadata",
    "x-metadata",
    "manifest",
    "spine",
    "tours",
    "guide",
]


class K8RESCProcessor(object):
    def __init__(self, data, debug=False):
        self._debug = debug
        self.resc = None
        self.opos = 0
        self.extrameta = []
        self.cover_name = None
        self.spine_idrefs = {}
        self.spine_order = []
        self.spine_pageattributes = {}
        self.spine_ppd = None
        # need3 indicate the book has fields which require epub3.
        # but the estimation of the source epub version from the fields is difficult.
        self.need3 = False
        self.package_ver = None
        self.extra_metadata = []
        self.refines_metadata = []
        self.extra_attributes = []
        # get header
        start_pos = data.find(b"<")
        self.resc_header = data[:start_pos]
        # get resc data length
        start = self.resc_header.find(b"=") + 1
        end = self.resc_header.find(b"&", start)
        resc_size = 0
        if end > 0:
            resc_size = fromBase32(self.resc_header[start:end])
        resc_rawbytes = len(data) - start_pos
        if resc_rawbytes == resc_size:
            self.resc_length = resc_size
        else:
            # Most RESC has a nul string at its tail but some do not.
            end_pos = data.find(b"\x00", start_pos)
            if end_pos < 0:
                self.resc_length = resc_rawbytes
            else:
                self.resc_length = end_pos - start_pos
        if self.resc_length != resc_size:
            logger.debug(
                "Warning: RESC section length({:d}bytes) does not match its size({:d}bytes).".format(
                    self.resc_length, resc_size
                )
            )
        # now parse RESC after converting it to unicode from utf-8
        self.resc = unicode_str(data[start_pos : start_pos + self.resc_length])
        self.parseData()

    def prepend_to_spine(self, key, idref, linear, properties):
        self.spine_order = [key] + self.spine_order
        self.spine_idrefs[key] = idref
        attributes = {}
        if linear is not None:
            attributes["linear"] = linear
        if properties is not None:
            attributes["properties"] = properties
        self.spine_pageattributes[key] = attributes

    # RESC tag iterator
    def resc_tag_iter(self):
        tcontent = last_tattr = None
        prefix = [""]
        while True:
            text, tag = self.parseresc()
            if text is None and tag is None:
                break
            if text is not None:
                tcontent = text.rstrip(" \r\n")
            else:  # we have a tag
                ttype, tname, tattr = self.parsetag(tag)
                if ttype == "begin":
                    tcontent = None
                    prefix.append(tname + ".")
                    if tname in _OPF_PARENT_TAGS:
                        yield "".join(prefix), tname, tattr, tcontent
                    else:
                        last_tattr = tattr
                else:  # single or end
                    if ttype == "end":
                        prefix.pop()
                        tattr = last_tattr
                        last_tattr = None
                        if tname in _OPF_PARENT_TAGS:
                            tname += "-end"
                    yield "".join(prefix), tname, tattr, tcontent
                    tcontent = None

    # now parse the RESC to extract spine and extra metadata info
    def parseData(self):
        for prefix, tname, tattr, tcontent in self.resc_tag_iter():
            if self._debug:
                logger.debug(
                    "   Parsing RESC: %s %s %s %s" % (prefix, tname, tattr, tcontent)
                )
            if tname == "package":
                self.package_ver = tattr.get("version", "2.0")
                package_prefix = tattr.get("prefix", "")
                if self.package_ver.startswith("3") or package_prefix.startswith(
                    "rendition"
                ):
                    self.need3 = True
            if tname == "spine":
                self.spine_ppd = tattr.get("page-progession-direction", None)
                if self.spine_ppd is not None and self.spine_ppd == "rtl":
                    self.need3 = True
            if tname == "itemref":
                skelid = tattr.pop("skelid", None)
                if skelid is None and len(self.spine_order) == 0:
                    # assume it was removed initial coverpage
                    skelid = "coverpage"
                    tattr["linear"] = "no"
                self.spine_order.append(skelid)
                idref = tattr.pop("idref", None)
                if idref is not None:
                    idref = "x_" + idref
                self.spine_idrefs[skelid] = idref
                if "id" in tattr:
                    del tattr["id"]
                # tattr["id"] = 'x_' + tattr["id"]
                if "properties" in tattr:
                    self.need3 = True
                self.spine_pageattributes[skelid] = tattr
            if tname == "meta" or tname.startswith("dc:"):
                if "refines" in tattr or "property" in tattr:
                    self.need3 = True
                if tattr.get("name", "") == "cover":
                    cover_name = tattr.get("content", None)
                    if cover_name is not None:
                        cover_name = "x_" + cover_name
                    self.cover_name = cover_name
                else:
                    self.extrameta.append([tname, tattr, tcontent])

    # parse and return either leading text or the next tag
    def parseresc(self):
        p = self.opos
        if p >= len(self.resc):
            return None, None
        if self.resc[p] != "<":
            res = self.resc.find("<", p)
            if res == -1:
                res = len(self.resc)
            self.opos = res
            return self.resc[p:res], None
        # handle comment as a special case
        if self.resc[p : p + 4] == "<!--":
            te = self.resc.find("-->", p + 1)
            if te != -1:
                te = te + 2
        else:
            te = self.resc.find(">", p + 1)
            ntb = self.resc.find("<", p + 1)
            if ntb != -1 and ntb < te:
                self.opos = ntb
                return self.resc[p:ntb], None
        self.opos = te + 1
        return None, self.resc[p : te + 1]

    # parses tag to identify:  [tname, ttype, tattr]
    #    tname: tag name
    #    ttype: tag type ('begin', 'end' or 'single');
    #    tattr: dictionary of tag atributes
    def parsetag(self, s):
        p = 1
        tname = None
        ttype = None
        tattr = dict_()
        while s[p : p + 1] == " ":
            p += 1
        if s[p : p + 1] == "/":
            ttype = "end"
            p += 1
            while s[p : p + 1] == " ":
                p += 1
        b = p
        while s[p : p + 1] not in (">", "/", " ", '"', "'", "\r", "\n"):
            p += 1
        tname = s[b:p].lower()
        # some special cases
        if tname == "?xml":
            tname = "xml"
        if tname == "!--":
            ttype = "single"
            comment = s[p:-3].strip()
            tattr["comment"] = comment
        if ttype is None:
            # parse any attributes of begin or single tags
            while s.find("=", p) != -1:
                while s[p : p + 1] == " ":
                    p += 1
                b = p
                while s[p : p + 1] != "=":
                    p += 1
                aname = s[b:p].lower()
                aname = aname.rstrip(" ")
                p += 1
                while s[p : p + 1] == " ":
                    p += 1
                if s[p : p + 1] in ('"', "'"):
                    p = p + 1
                    b = p
                    while s[p : p + 1] not in ('"', "'"):
                        p += 1
                    val = s[b:p]
                    p += 1
                else:
                    b = p
                    while s[p : p + 1] not in (">", "/", " "):
                        p += 1
                    val = s[b:p]
                tattr[aname] = val
        if ttype is None:
            ttype = "begin"
            if s.find("/", p) >= 0:
                ttype = "single"
        return ttype, tname, tattr

    def taginfo_toxml(self, taginfo):
        res = []
        tname, tattr, tcontent = taginfo
        res.append("<" + tname)
        if tattr is not None:
            for key in tattr:
                res.append(" " + key + '="' + tattr[key] + '"')
        if tcontent is not None:
            res.append(">" + tcontent + "</" + tname + ">\n")
        else:
            res.append("/>\n")
        return "".join(res)

    def hasSpine(self):
        return len(self.spine_order) > 0

    def needEPUB3(self):
        return self.need3

    def hasRefines(self):
        for [tname, tattr, tcontent] in self.extrameta:
            if "refines" in tattr:
                return True
        return False

    def createMetadata(self, epubver):
        for taginfo in self.extrameta:
            tname, tattr, tcontent = taginfo
            if "refines" in tattr:
                if epubver == "F" and "property" in tattr:
                    attr = ' id="%s" opf:%s="%s"\n' % (
                        tattr["refines"],
                        tattr["property"],
                        tcontent,
                    )
                    self.extra_attributes.append(attr)
                else:
                    tag = self.taginfo_toxml(taginfo)
                    self.refines_metadata.append(tag)
            else:
                tag = self.taginfo_toxml(taginfo)
                self.extra_metadata.append(tag)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim:ts=4:sw=4:softtabstop=4:smarttab:expandtab

from __future__ import unicode_literals, division, absolute_import, print_function

import os
from .unipath import pathof
from loguru import logger


import re

# note: re requites the pattern to be the exact same type as the data to be searched in python3
# but u"" is not allowed for the pattern itself only b""

from .mobi_utils import toBase32
from .mobi_index import MobiIndex

DEBUG_NCX = False


class ncxExtract:
    def __init__(self, mh, files):
        self.mh = mh
        self.sect = self.mh.sect
        self.files = files
        self.isNCX = False
        self.mi = MobiIndex(self.sect)
        self.ncxidx = self.mh.ncxidx
        self.indx_data = None

    def parseNCX(self):
        indx_data = []
        tag_fieldname_map = {
            1: ["pos", 0],
            2: ["len", 0],
            3: ["noffs", 0],
            4: ["hlvl", 0],
            5: ["koffs", 0],
            6: ["pos_fid", 0],
            21: ["parent", 0],
            22: ["child1", 0],
            23: ["childn", 0],
        }
        if self.ncxidx != 0xFFFFFFFF:
            outtbl, ctoc_text = self.mi.getIndexData(self.ncxidx, "NCX")
            if DEBUG_NCX:
                logger.debug(ctoc_text)
                logger.debug(outtbl)
            num = 0
            for [text, tagMap] in outtbl:
                tmp = {
                    "name": text.decode("utf-8"),
                    "pos": -1,
                    "len": 0,
                    "noffs": -1,
                    "text": "Unknown Text",
                    "hlvl": -1,
                    "kind": "Unknown Kind",
                    "pos_fid": None,
                    "parent": -1,
                    "child1": -1,
                    "childn": -1,
                    "num": num,
                }
                for tag in tag_fieldname_map:
                    [fieldname, i] = tag_fieldname_map[tag]
                    if tag in tagMap:
                        fieldvalue = tagMap[tag][i]
                        if tag == 6:
                            pos_fid = toBase32(fieldvalue, 4).decode("utf-8")
                            fieldvalue2 = tagMap[tag][i + 1]
                            pos_off = toBase32(fieldvalue2, 10).decode("utf-8")
                            fieldvalue = "kindle:pos:fid:%s:off:%s" % (pos_fid, pos_off)
                        tmp[fieldname] = fieldvalue
                        if tag == 3:
                            toctext = ctoc_text.get(fieldvalue, "Unknown Text")
                            toctext = toctext.decode(self.mh.codec)
                            tmp["text"] = toctext
                        if tag == 5:
                            kindtext = ctoc_text.get(fieldvalue, "Unknown Kind")
                            kindtext = kindtext.decode(self.mh.codec)
                            tmp["kind"] = kindtext
                indx_data.append(tmp)
                if DEBUG_NCX:
                    print("record number: ", num)
                    print(
                        "name: ", tmp["name"],
                    )
                    print("position", tmp["pos"], " length: ", tmp["len"])
                    print("text: ", tmp["text"])
                    print("kind: ", tmp["kind"])
                    print("heading level: ", tmp["hlvl"])
                    print("parent:", tmp["parent"])
                    print(
                        "first child: ", tmp["child1"], " last child: ", tmp["childn"]
                    )
                    print("pos_fid is ", tmp["pos_fid"])
                    print("\n\n")
                num += 1
        self.indx_data = indx_data
        return indx_data

    def buildNCX(self, htmlfile, title, ident, lang):
        indx_data = self.indx_data

        ncx_header = """<?xml version='1.0' encoding='utf-8'?>
<ncx xmlns="http://www.daisy.org/z3986/2005/ncx/" version="2005-1" xml:lang="%s">
<head>
<meta content="%s" name="dtb:uid"/>
<meta content="%d" name="dtb:depth"/>
<meta content="mobiunpack.py" name="dtb:generator"/>
<meta content="0" name="dtb:totalPageCount"/>
<meta content="0" name="dtb:maxPageNumber"/>
</head>
<docTitle>
<text>%s</text>
</docTitle>
<navMap>
"""

        ncx_footer = """  </navMap>
</ncx>
"""

        ncx_entry = """<navPoint id="%s" playOrder="%d">
<navLabel>
<text>%s</text>
</navLabel>
<content src="%s"/>"""

        # recursive part
        def recursINDX(max_lvl=0, num=0, lvl=0, start=-1, end=-1):
            if start > len(indx_data) or end > len(indx_data):
                print("Warning: missing INDX child entries", start, end, len(indx_data))
                return ""
            if DEBUG_NCX:
                logger.debug("recursINDX lvl %d from %d to %d" % (lvl, start, end))
            xml = ""
            if start <= 0:
                start = 0
            if end <= 0:
                end = len(indx_data)
            if lvl > max_lvl:
                max_lvl = lvl
            indent = "  " * (2 + lvl)

            for i in range(start, end):
                e = indx_data[i]
                if not e["hlvl"] == lvl:
                    continue
                # open entry
                num += 1
                link = "%s#filepos%d" % (htmlfile, e["pos"])
                tagid = "np_%d" % num
                entry = ncx_entry % (tagid, num, e["text"], link)
                entry = re.sub(re.compile("^", re.M), indent, entry, 0)
                xml += entry + "\n"
                # recurs
                if e["child1"] >= 0:
                    xmlrec, max_lvl, num = recursINDX(
                        max_lvl, num, lvl + 1, e["child1"], e["childn"] + 1
                    )
                    xml += xmlrec
                # close entry
                xml += indent + "</navPoint>\n"
            return xml, max_lvl, num

        body, max_lvl, num = recursINDX()
        header = ncx_header % (lang, ident, max_lvl + 1, title)
        ncx = header + body + ncx_footer
        if not len(indx_data) == num:
            print("Warning: different number of entries in NCX", len(indx_data), num)
        return ncx

    def writeNCX(self, metadata):
        # build the xml
        self.isNCX = True
        logger.debug("Write ncx")
        # htmlname = os.path.basename(self.files.outbase)
        # htmlname += '.html'
        htmlname = "book.html"
        xml = self.buildNCX(
            htmlname,
            metadata["Title"][0],
            metadata["UniqueID"][0],
            metadata.get("Language")[0],
        )
        # write the ncx file
        # ncxname = os.path.join(self.files.mobi7dir, self.files.getInputFileBasename() + '.ncx')
        ncxname = os.path.join(self.files.mobi7dir, "toc.ncx")
        with open(pathof(ncxname), "wb") as f:
            f.write(xml.encode("utf-8"))

    def buildK8NCX(self, indx_data, title, ident, lang):
        ncx_header = """<?xml version='1.0' encoding='utf-8'?>
<ncx xmlns="http://www.daisy.org/z3986/2005/ncx/" version="2005-1" xml:lang="%s">
<head>
<meta content="%s" name="dtb:uid"/>
<meta content="%d" name="dtb:depth"/>
<meta content="mobiunpack.py" name="dtb:generator"/>
<meta content="0" name="dtb:totalPageCount"/>
<meta content="0" name="dtb:maxPageNumber"/>
</head>
<docTitle>
<text>%s</text>
</docTitle>
<navMap>
"""

        ncx_footer = """  </navMap>
</ncx>
"""

        ncx_entry = """<navPoint id="%s" playOrder="%d">
<navLabel>
<text>%s</text>
</navLabel>
<content src="%s"/>"""

        # recursive part
        def recursINDX(max_lvl=0, num=0, lvl=0, start=-1, end=-1):
            if start > len(indx_data) or end > len(indx_data):
                print("Warning: missing INDX child entries", start, end, len(indx_data))
                return ""
            if DEBUG_NCX:
                logger.debug("recursINDX lvl %d from %d to %d" % (lvl, start, end))
            xml = ""
            if start <= 0:
                start = 0
            if end <= 0:
                end = len(indx_data)
            if lvl > max_lvl:
                max_lvl = lvl
            indent = "  " * (2 + lvl)

            for i in range(start, end):
                e = indx_data[i]
                htmlfile = e["filename"]
                desttag = e["idtag"]
                if not e["hlvl"] == lvl:
                    continue
                # open entry
                num += 1
                if desttag == "":
                    link = "Text/%s" % htmlfile
                else:
                    link = "Text/%s#%s" % (htmlfile, desttag)
                tagid = "np_%d" % num
                entry = ncx_entry % (tagid, num, e["text"], link)
                entry = re.sub(re.compile("^", re.M), indent, entry, 0)
                xml += entry + "\n"
                # recurs
                if e["child1"] >= 0:
                    xmlrec, max_lvl, num = recursINDX(
                        max_lvl, num, lvl + 1, e["child1"], e["childn"] + 1
                    )
                    xml += xmlrec
                # close entry
                xml += indent + "</navPoint>\n"
            return xml, max_lvl, num

        body, max_lvl, num = recursINDX()
        header = ncx_header % (lang, ident, max_lvl + 1, title)
        ncx = header + body + ncx_footer
        if not len(indx_data) == num:
            print("Warning: different number of entries in NCX", len(indx_data), num)
        return ncx

    def writeK8NCX(self, ncx_data, metadata):
        # build the xml
        self.isNCX = True
        logger.debug("Write K8 ncx")
        xml = self.buildK8NCX(
            ncx_data,
            metadata["Title"][0],
            metadata["UniqueID"][0],
            metadata.get("Language")[0],
        )
        bname = "toc.ncx"
        ncxname = os.path.join(self.files.k8oebps, bname)
        with open(pathof(ncxname), "wb") as f:
            f.write(xml.encode("utf-8"))
